import logging
from abc import ABC, abstractmethod
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.tag import Tag, TagTarget

logger = logging.getLogger("picsellia")


class Taggable(ABC):
    def __init__(self, target_type: TagTarget) -> None:
        self._target_type = target_type

    @property
    def target_type(self) -> TagTarget:
        return self._target_type

    @property
    @abstractmethod
    def ids(self) -> list[UUID]:
        pass

    @property
    @abstractmethod
    def connexion(self) -> Connection:
        pass

    @exception_handler
    @beartype
    def add_tags(self, tags: Tag | list[Tag]) -> None:
        """Add some tags to an object.
        It can be used on Data/MultiData/Asset/MultiAsset/DatasetVersion/Dataset/Model/ModelVersion.

        You can give a Tag or a list of (Tag).

        Examples:
            ```python
            tag_bicycle = client.create_tag("bicycle", Target.DATA)
            tag_car = client.create_tag("car", Target.DATA)
            tag_truck = client.create_tag("truck", Target.DATA)

            data.add_tags(tag_bicycle)
            data.add_tags([tag_car, tag_truck])
            ```
        """
        if isinstance(tags, Tag):
            tags = [tags]

        assert tags != [], "Given tags are empty. They can't be empty"

        for tag in tags:
            assert (
                tag.target_type == self.target_type
            ), f"Given tag ({tag.name}) is targeted on {tag.target_type}. It can't be added because on a {self.target_type} "

        for tag in tags:
            payload = self.ids
            self.connexion.post(
                f"/api/tag/{tag.id}/attach",
                data=orjson.dumps(payload),
            )
        logger.info(f"{len(tags)} tags added to {self}).")

    @exception_handler
    @beartype
    def remove_tags(self, tags: Tag | list[Tag]) -> None:
        """Remove some tags from an object (can be used on Data/Asset/DatasetVersion/Dataset/Model/ModelVersion)

        You can give a (Tag) or a list of (Tag).

        Examples:
            ```python
            data.remove_tags(tag_bicycle)
            data.remove_tags([tag_car, tag_truck])
            ```
        """
        if isinstance(tags, Tag):
            tags = [tags]

        assert tags != [], "Given tags are empty. They can't be empty"

        for tag in tags:
            assert (
                tag.target_type == self._target_type
            ), f"Given tag ({tag.name}) can't be removed because it is not targeted on data"

        for tag in tags:
            payload = self.ids
            self.connexion.post(
                f"/api/tag/{tag.id}/detach",
                data=orjson.dumps(payload),
            )
        logger.info(f"{len(tags)} tags removed from {self}.")



import abc
import logging
import os
from collections.abc import Iterable, Mapping
from io import IOBase
from pathlib import Path
from typing import IO, Any
from urllib.parse import urljoin
from uuid import UUID

import orjson
import requests
from beartype import beartype
from requests import Response, Session
from requests.exceptions import ConnectionError, InvalidJSONError

import picsellia.exceptions as exceptions
from picsellia import __version__
from picsellia.decorators import exception_handler, retry
from picsellia.services.upload.file import FileUploader
from picsellia.types.enums import ObjectDataType

logger = logging.getLogger("picsellia")

DEFAULT_TIMEOUT = 30

DATA_TYPE = str | bytes | Mapping[str, Any] | Iterable[tuple[str, str, None]] | IO


def reset_body(body):
    if "files" in body and isinstance(body["files"], dict):
        logger.debug("Seeking buffered readers to zero")
        for value in body["files"].values():
            if isinstance(value, IOBase):
                try:
                    value.seek(0)
                except Exception:
                    pass


def wrapped_request(f):
    def decorated(self, *args, regenerate_jwt=True, **kwargs):
        resp = f(self, *args, **kwargs)

        if resp.status_code == 401:
            if regenerate_jwt:
                logger.info(f"Regenerating connection token to {self.host}...")
                self._jwt, self._expires_in = self.generate_jwt()
                reset_body(kwargs)
                return decorated(self, *args, regenerate_jwt=False, **kwargs)
            else:
                raise exceptions.UnauthorizedError(
                    "You are not authorized to do this: regeneration of connection token failed."
                )

        return resp

    return decorated


def handle_response(f):
    def decorated(*args, **kwargs):
        response = f(*args, **kwargs)
        check_status_code(response)
        return response

    return decorated


# No exception handling: exception needs to be raised
def check_status_code(response: Response):  # noqa: C901
    status = int(response.status_code)
    if status == 200:
        logger.debug("OK.")
    elif status == 201:
        logger.debug("Resource created.")
    elif status == 202:
        logger.debug("Accepted.")
    elif status == 203:
        logger.debug("OK.")
    elif status == 204:
        logger.debug("No content.")
    elif status != 208 and status <= 299:
        logger.debug(f"Request done : {status}")
    else:
        try:
            data = response.json()
            if "message" not in data:
                data["message"] = f"No message ({status})"

            message = data["message"]
            if "detail" in data and data["detail"] is not None and data["detail"] != []:
                if isinstance(data["detail"], list):
                    message += ". Detail: \n"
                    for item in data["detail"]:
                        message += " > " + str(item) + "\n"
                else:
                    message = message + ". Detail: " + str(data["detail"])
        except (KeyError, InvalidJSONError):
            message = response.text

        logger.debug(
            f"Platform returned an error (status code: {status}). Message : {message}"
        )

        if status == 208:
            raise exceptions.DistantStorageError(
                "An object has already this name in S3."
            )
        if status == 400:
            raise exceptions.BadRequestError(message)
        if status == 401:
            raise exceptions.UnauthorizedError(message)
        if status == 402:
            raise exceptions.InsufficientResourcesError(message)
        if status == 403:
            raise exceptions.ForbiddenError(message)
        if status == 404:
            raise exceptions.ResourceNotFoundError(message)
        if status == 405:
            raise exceptions.PicselliaError(f"Method not allowed: {message}")
        if status == 409:
            raise exceptions.ResourceConflictError(message)
        if status == 413:
            raise exceptions.RequestTooLargeError(
                "There is too much data in your request."
            )
        if status == 422:
            raise exceptions.BadRequestError(message)
        if status == 423:
            raise exceptions.ResourceLockedError(message)
        if status == 429:
            raise exceptions.TooManyRequestError(message)
        if status == 500:
            raise exceptions.InternalServerError(message)
        if status == 502:
            raise exceptions.BadGatewayError("Picsellia is unavailable at the moment.")
        raise exceptions.PicselliaError(f"[{status}] Something went wrong, {message}")


class AbstractConnection(abc.ABC):
    def __init__(self, host: str, session: Session | None = None) -> None:
        self.host = host
        if session is not None:
            self.session = session
        else:
            self.session = requests.Session()

    def __enter__(self):
        return self

    def __exit__(self, exc_type, exc_val, exc_tb):
        if self.session is not None:
            self.session.close()

    @property
    @abc.abstractmethod
    def headers(self):
        pass

    def get(self, path: str, params: dict | None = None, stream: bool = False):
        url = urljoin(self.host, path)
        return self.session.get(
            url=url, headers=self.headers, params=params, stream=stream
        )

    def xget(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        stream: bool = False,
    ):
        url = urljoin(self.host, path)
        return self.session.request(
            method="XGET",
            url=url,
            data=data,
            headers=self.headers,
            params=params,
            stream=stream,
        )

    def post(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        files: Any | None = None,
    ):
        url = urljoin(self.host, path)
        return self.session.post(
            url=url, data=data, headers=self.headers, params=params, files=files
        )

    def put(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        url = urljoin(self.host, path)
        return self.session.put(url=url, data=data, headers=self.headers, params=params)

    def patch(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        url = urljoin(self.host, path)
        return self.session.patch(
            url=url, data=data, headers=self.headers, params=params
        )

    def delete(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        url = urljoin(self.host, path)
        return self.session.delete(
            url=url, data=data, headers=self.headers, params=params
        )


class Connection(AbstractConnection):
    def __init__(
        self,
        host: str,
        api_token: str,
        content_type: str = "application/json",
        session: Session | None = None,
    ) -> None:
        super().__init__(host, session)
        self.api_token = api_token
        self._headers = {
            "Authorization": "Bearer " + api_token,
            "User-Agent": f"Picsellia-SDK/{__version__}",
        }
        if content_type:
            self._headers["Content-type"] = content_type

        self._connector_id = None
        self._organization_id = None

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, Connection):
            return self.host == __o.host and self.api_token == __o.api_token

        return False

    @property
    def headers(self):
        return self._headers

    @property
    def connector_id(self):
        if self._connector_id is None:
            raise exceptions.NoConnectorFound(
                "This organization has no default connector, and connect retrieve and upload files."
            )
        return self._connector_id

    @connector_id.setter
    def connector_id(self, value):
        self._connector_id = value

    @property
    def organization_id(self):
        return self._organization_id

    @organization_id.setter
    def organization_id(self, value):
        self._organization_id = value
        self.headers["X-Picsellia-Organization"] = str(self._organization_id)

    @handle_response
    @retry(requests.ConnectionError)
    def get(self, path: str, params: dict | None = None, stream: bool = False):
        return super().get(path, params, stream)

    @handle_response
    @retry(requests.ConnectionError)
    def xget(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        stream: bool = False,
    ):
        return super().xget(path, data, params, stream)

    @handle_response
    @retry(requests.ConnectionError)
    def post(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        files: Any | None = None,
    ):
        return super().post(path, data, params, files)

    @handle_response
    @retry(requests.ConnectionError)
    def put(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().put(path, data, params)

    @handle_response
    @retry(requests.ConnectionError)
    def patch(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().patch(path, data, params)

    @handle_response
    @retry(requests.ConnectionError)
    def delete(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().delete(path, data, params)

    ##############################################################
    # ------------------------- UPLOAD ------------------------- #
    ##############################################################
    @exception_handler
    @beartype
    def _generate_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        connector_id: UUID | None = None,
        context: dict[str, UUID] | None = None,
        upload_dir: str | None = None,
    ) -> str:
        if connector_id is None:
            connector_id = self.connector_id

        payload: dict[str, Any] = {"filename": filename, "type": object_name_type}
        if context:
            payload["context"] = context

        if upload_dir:
            payload["upload_dir"] = upload_dir

        r = self.post(
            path=f"/api/organization/{self.organization_id}/connector/{connector_id}/generate_object_name",
            data=orjson.dumps(payload),
        ).json()
        return r["object_name"]

    @beartype
    def generate_data_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        connector_id: UUID | None = None,
        upload_dir: str | None = None,
    ):
        if object_name_type not in [
            ObjectDataType.DATA,
            ObjectDataType.DATA_PROJECTION,
        ]:
            raise RuntimeError(
                f"Cannot generate data object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename, object_name_type, connector_id, upload_dir=upload_dir
        )

    @beartype
    def generate_dataset_version_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        dataset_version_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type not in [ObjectDataType.CAMPAIGN_FILE]:
            raise RuntimeError(
                f"Cannot generate dataset version object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id,
            context={"dataset_version_id": dataset_version_id},
        )

    @beartype
    def generate_deployment_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        deployment_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type not in [ObjectDataType.REVIEW_CAMPAIGN_FILE]:
            raise RuntimeError(
                f"Cannot generate deployment object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id,
            context={"deployment_id": deployment_id},
        )

    @beartype
    def generate_job_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        job_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type not in [ObjectDataType.LOGGING]:
            raise RuntimeError(
                f"Cannot generate job object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"job_id": job_id},
        )

    @beartype
    def generate_experiment_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        experiment_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type not in (
            ObjectDataType.ARTIFACT,
            ObjectDataType.LOG_IMAGE,
            ObjectDataType.LOGGING,
        ):
            raise RuntimeError(
                f"Cannot generate experiment object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"experiment_id": experiment_id},
        )

    @beartype
    def generate_model_version_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        model_version_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type not in (
            ObjectDataType.MODEL_THUMB,
            ObjectDataType.MODEL_FILE,
        ):
            raise RuntimeError(
                f"Cannot generate model version object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"model_version_id": model_version_id},
        )

    @beartype
    def generate_report_object_name(
        self,
        filename: str,
        object_name_type: ObjectDataType,
        dataset_version_id: UUID,
        report_id: UUID,
        connector_id: UUID | None = None,
    ):
        if object_name_type != ObjectDataType.AGENTS_REPORT:
            raise RuntimeError(
                f"Cannot generate report object name with type {object_name_type}"
            )
        return self._generate_object_name(
            filename,
            object_name_type,
            connector_id=connector_id,
            context={"report_id": report_id, "dataset_version_id": dataset_version_id},
        )

    def upload_file(
        self,
        object_name: str,
        path: str | Path,
        connector_id: UUID | None = None,
    ) -> tuple[requests.Response, bool, str]:
        """Upload a single file to the server.
        If file is bigger than 5Mb, it will send it by multipart.

        Arguments:
            path (str): Absolute path to the file
            object_name (str): Destination object name.
            connector_id (UUID): Connector on which you need to upload file, if it's not default connector.
        """
        if connector_id is None:
            connector_id = self.connector_id

        uploader = FileUploader(connector_id, self.session, self.host, self.headers)
        return uploader.upload(object_name, path)

    ##############################################################
    # ------------------------ DOWNLOAD ------------------------ #
    ##############################################################
    def init_download(self, object_name: str, connector_id: UUID | None = None) -> str:
        """Retrieve a presigned url of this object name in order to download it"""
        if connector_id is None:
            connector_id = self.connector_id

        payload = {"object_name": object_name}
        r = self.post(
            path=f"/api/object-storage/{connector_id}/retrieve_presigned_url",
            data=orjson.dumps(payload),
        )

        if r.status_code != 200:
            raise exceptions.DistantStorageError("Errors while getting a presigned url")

        r = r.json()
        if "presigned_url" not in r:
            raise exceptions.DistantStorageError(
                "Errors while getting a presigned url. Unparsable response"
            )

        return r["presigned_url"]

    def do_download_file(
        self,
        path: str | Path,
        url: str,
        is_large: bool,
        force_replace: bool,
        retry_count: int = 1,
    ) -> bool:
        try:
            return self._do_download_file(path, url, is_large, force_replace)
        except (exceptions.NetworkError, ConnectionError) as e:
            # Here for retro compatibility
            raise exceptions.DownloadError(
                f"Could not download {url} into {path}"
            ) from e

    @retry((exceptions.NetworkError, ConnectionError))
    def _do_download_file(
        self,
        path: str | Path,
        url: str,
        is_large: bool,
        force_replace: bool,
    ) -> bool:
        """Retrieve a presigned url of this object name in order to download it"""
        if os.path.exists(path) and not force_replace:
            return False

        parent_path = Path(path).parent.absolute()
        os.makedirs(parent_path, exist_ok=True)

        response = self.session.get(url, stream=is_large, timeout=DEFAULT_TIMEOUT)

        if response.status_code == 429 or (500 <= response.status_code < 600):
            raise exceptions.NetworkError(
                f"Response status code is {response.status_code}. Could not get {url}"
            )

        response.raise_for_status()

        total_length = response.headers.get("content-length")
        if total_length is None:
            raise exceptions.NetworkError(
                "Downloaded content is empty but response is 200"
            )

        with open(path, "wb") as handler:
            if not is_large:
                handler.write(response.content)
            else:
                for data in response.iter_content(chunk_size=4096):
                    handler.write(data)

        return True


class JwtServiceConnection(AbstractConnection):
    def __init__(
        self,
        host: str,
        jwt_identifier_data: dict,
        login_path: str,
        session: Session | None = None,
    ) -> None:
        """JwtServiceConnection may be used to handle a JwtAuthentication connexion with a service.
        You need to give a dict `jwt_identifier_data` that will be sent to the host to ensures validity of your request.
        This data will depend on service contacted. For example for a Deployment / Jwt request on Oracle, this will be :
        {
            "api_token" : <user_api_token>,
            "deployment_id" : <deployment_id>
        }
        """
        super().__init__(host, session)
        self.jwt_identifier_data = jwt_identifier_data
        self.login_path = login_path
        self._jwt, self._expires_in = None, None

    @property
    def jwt(self) -> str:
        if self._jwt is None:
            self._jwt, self._expires_in = self.generate_jwt()
        return self._jwt

    @property
    def headers(self):
        return {"Authorization": f"Bearer {self.jwt}"}

    def generate_jwt(self):
        url = urljoin(self.host, self.login_path)
        response = self.session.post(
            url=url, data=orjson.dumps(self.jwt_identifier_data)
        )

        if response.status_code != 200:
            raise exceptions.UnauthorizedError("Unauthorized attempt to connect.")

        try:
            data = response.json()
            return data["jwt"], data["expires"]
        except Exception:
            raise exceptions.UnauthorizedError(
                "Cannot parse response from external service. Please contact support."
            )

    @wrapped_request
    def get(self, path: str, params: dict | None = None, stream: bool = False):
        return super().get(path, params, stream)

    @wrapped_request
    def xget(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        stream: bool = False,
    ):
        return super().xget(path, data, params, stream)

    @wrapped_request
    def post(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
        files: Any | None = None,
    ):
        return super().post(path, data, params, files)

    @wrapped_request
    def put(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().put(path, data, params)

    @wrapped_request
    def patch(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().patch(path, data, params)

    @wrapped_request
    def delete(
        self,
        path: str,
        data: DATA_TYPE | None = None,
        params: dict | None = None,
    ):
        return super().delete(path, data, params)


Connexion = Connection
JwtServiceConnexion = JwtServiceConnection



import base64
import json
import logging
import mimetypes
import os
from datetime import date
from pathlib import Path
from typing import Any
from uuid import UUID

import orjson
import requests
from beartype import beartype
from deprecation import deprecated

from picsellia.colors import Colors
from picsellia.decorators import exception_handler, retry
from picsellia.exceptions import (
    BadConfigurationContinuousTrainingError,
    BadRequestError,
    ContentTypeUnknown,
    MonitorError,
    MonitoringConnectionError,
    NoDataError,
    NoShadowModel,
    PicselliaError,
    PredictionError,
    ResourceNotFoundError,
)
from picsellia.sdk.connexion import Connection, JwtServiceConnection
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data
from picsellia.sdk.datalake import Datalake
from picsellia.sdk.dataset import DatasetVersion
from picsellia.sdk.datasource import DataSource
from picsellia.sdk.model_version import ModelVersion
from picsellia.sdk.predicted_asset import MultiPredictedAsset, PredictedAsset
from picsellia.sdk.project import Project
from picsellia.sdk.resource import Resource
from picsellia.sdk.review_campaign import ReviewCampaign
from picsellia.sdk.tag import Tag
from picsellia.sdk.taggable import Taggable
from picsellia.sdk.worker import Worker
from picsellia.services.lister.predicted_asset_lister import (
    PredictedAssetFilter,
    PredictedAssetLister,
)
from picsellia.types.enums import (
    ContinuousDeploymentPolicy,
    ContinuousTrainingTrigger,
    ContinuousTrainingType,
    InferenceType,
    ObjectDataType,
    ServiceMetrics,
    SupportedContentType,
    TagTarget,
    WorkerType,
)
from picsellia.types.schemas import DeploymentSchema, TargetDatalakeConnectorSchema
from picsellia.types.schemas_prediction import PredictionFormat

logger = logging.getLogger("picsellia")


class Deployment(Dao, Taggable, Resource):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)
        Taggable.__init__(self, TagTarget.DEPLOYMENT)
        Resource.__init__(self, "deployment")

        # lazy loading of target datalake connector
        self._target_datalake_connector = None

        deployment = self.refresh(data)
        if deployment.oracle_host is not None:
            try:
                self._oracle_connexion = JwtServiceConnection(
                    deployment.oracle_host,
                    {
                        "api_token": self.connexion.api_token,
                        "deployment_id": str(self.id),
                    },
                    login_path="/api/auth/login",
                )
                if self._oracle_connexion.jwt is None:  # pragma: no cover
                    raise PicselliaError("Cannot authenticate to oracle")

                logging.info(
                    f"Connected with monitoring service at {deployment.oracle_host}"
                )
            except Exception as e:
                logger.error(
                    f"Could not bind {self} with our monitoring service at {deployment.oracle_host} because : {e}"
                )
                self._oracle_connexion.session.close()
                self._oracle_connexion = None
        else:  # pragma: no cover
            self._oracle_connexion = None

        if deployment.serving_host is not None:
            try:
                self._serving_connexion = JwtServiceConnection(
                    deployment.serving_host,
                    {
                        "api_token": self.connexion.api_token,
                        "deployment_id": str(self.id),
                    },
                    login_path="/api/login",
                )
                if self._serving_connexion.jwt is None:  # pragma: no cover
                    raise PicselliaError("Cannot authenticate to serving")
                logging.info(
                    f"Connected with serving service at {deployment.serving_host}"
                )
            except Exception as e:
                logger.error(
                    f"Could not bind {self} with our serving service at {deployment.serving_host} because : {e}"
                )
                self._serving_connexion.session.close()
                self._serving_connexion = None
        else:  # pragma: no cover
            self._serving_connexion = None

    @property
    def name(self) -> str:
        """Name of this (Deployment)"""
        return self._name

    @property
    def type(self) -> InferenceType:
        """Type of this (Deployment)"""
        return self._type

    @property
    def oracle_connexion(self) -> JwtServiceConnection:
        assert (
            self._oracle_connexion is not None
        ), "You can't use this function with this deployment. Please contact the support."
        return self._oracle_connexion

    @property
    def serving_connexion(self) -> JwtServiceConnection:
        assert (
            self._serving_connexion is not None
        ), "You can't use this function with this deployment. Please contact the support."
        return self._serving_connexion

    def __str__(self):
        return f"{Colors.CYAN}Deployment '{self.name}' {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> DeploymentSchema:
        schema = DeploymentSchema(**data)
        self._name = schema.name
        self._type = schema.type
        self._review_campaign_id = schema.review_campaign_id
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/deployment/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def get_tags(self) -> list[Tag]:
        """Retrieve the tags of your deployment.

        Examples:
            ```python
            tags = deployment.get_tags()
            assert tags[0].name == "cool"
            ```

        Returns:
            A list of (Tag) objects
        """
        r = self.sync()
        return [Tag(self.connexion, item) for item in r["tags"]]

    @exception_handler
    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def retrieve_information(self) -> dict:
        """Retrieve some information about this deployment from service.

        Examples:
            ```python
            my_deployment.retrieve_information()
            ```
        Returns:
            A dict with information about this deployment
        """
        response = self.oracle_connexion.get(path=f"/api/deployment/{self.id}")
        if response.status_code > 500:
            raise MonitoringConnectionError()
        return response.json()

    @exception_handler
    @beartype
    def update(
        self,
        name: str | None = None,
        target_datalake: Datalake | None = None,
        min_threshold: float | None = None,
        sample_rate: float | None = None,
    ) -> None:
        """Update this deployment with a new name, another target datalake or a minimum threshold

        Examples:
            ```python
            a_tag.update(name="new name", min_threshold=0.4)
            ```
        Arguments:
            name (str, optional): New name of the deployment
            target_datalake (Datalake, optional): Datalake where data will be uploaded on new prediction
            min_threshold (float, optional): Minimum confidence threshold.
                    Serving will filter detection boxes or masks that have a detection score lower than this threshold
            sample_rate (float, optional): If less than 1, monitoring service will not send all images to platform,
                    but will still compute metrics.
        """
        payload = {}
        if name is not None:
            payload["name"] = name

        if min_threshold is not None:
            payload["min_threshold"] = min_threshold

        if target_datalake is not None:
            payload["target_datalake_id"] = target_datalake.id

        if sample_rate is not None:
            payload["sample_rate"] = sample_rate

        r = self.connexion.patch(
            f"/api/deployment/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self, force_delete: bool = False) -> None:
        self.connexion.delete(
            f"/api/deployment/{self.id}", params={"force_delete": force_delete}
        )
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def set_model(self, model_version: ModelVersion) -> None:
        """Set the model version to use for this deployment

        Examples:
            ```python
            model_version = client.get_model("my-model").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.set_model(model_version)
            ```
        Arguments:
            model_version (ModelVersion): a (ModelVersion) to use
        """
        payload = {"model_version_id": model_version.id}

        self.connexion.post(
            f"/api/deployment/{self.id}/model", data=orjson.dumps(payload)
        ).json()
        logger.info(f"{self} model is now {model_version}")

    @exception_handler
    @beartype
    def get_model_version(self) -> ModelVersion:
        """Retrieve currently used model version

        Examples:
            ```python
            model_version = deployment.get_model_version()
            ```

        Returns:
            A (ModelVersion) object
        """
        r = self.sync()

        r = self.connexion.get(f"/api/model/version/{r['model_version_id']}").json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def set_shadow_model(self, shadow_model_version: ModelVersion) -> None:
        """Set the shadow model version to use for this deployment

        Examples:
            ```python
            shadow_model_version = client.get_model("my-model").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.set_shadow_model(shadow_model_version)
            ```

        Arguments:
            shadow_model_version (ModelVersion): a (ModelVersion) to use
        """
        payload = {"model_version_id": shadow_model_version.id}

        self.connexion.post(
            f"/api/deployment/{self.id}/shadow", data=orjson.dumps(payload)
        ).json()
        logger.info(f"{self} shadow model is now {shadow_model_version}")

    @exception_handler
    @beartype
    def get_shadow_model(self) -> ModelVersion:
        """Retrieve currently used shadow model version

        Examples:
            ```python
            shadow_model = deployment.get_shadow_model()
            ```

        Returns:
            A (ModelVersion) object
        """
        r = self.sync()
        if "shadow_model_version_id" not in r or r["shadow_model_version_id"] is None:
            raise NoShadowModel("This deployment has no shadow model")

        r = self.connexion.get(
            f"/api/model/version/{r['shadow_model_version_id']}"
        ).json()
        return ModelVersion(self.connexion, r)

    @exception_handler
    @beartype
    def predict(
        self,
        file_path: str | Path,
        tags: str | Tag | list[Tag | str] | None = None,
        source: str | DataSource | None = None,
        metadata: dict | None = None,
        monitor: bool = True,
        upload_dir: str | None = None,
        custom_metadata: dict | None = None,
    ) -> dict:
        """Run a prediction on our Serving platform

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.predict('image_420.png', tags=["gonna", "give"], source="camera-1")
            ```
        Arguments:
            file_path (str or Path): path to the image to predict.
            tags (str, (Tag), list[str | Tag], optional): a list of tag to add to the data that will be created on the platform.
            source (str or DataSource, optional): a source to attach to the data that will be created on the platform.
            metadata (dict, optional): metadata to attach to the data that will be created on the platform.
            monitor (bool, optional): if True, will send prediction on Picsellia and our monitoring service. Defaults to True.
            upload_dir (str, optional): This parameter can only be used with private object-storages. Specify this parameter to prefix the object name of the data. Filename will still contain a generated uuid4
            custom_metadata (dict, optional): custom_metadata to attach to the data that will be created on the platform.

        Returns:
            A (dict) with information of the prediction
        """
        with open(file_path, "rb") as file:
            file_data = file.read()
            filename = Path(file_path).name
            return self.predict_bytes(
                filename=filename,
                raw_image=file_data,
                tags=tags,
                source=source,
                metadata=metadata,
                monitor=monitor,
                upload_dir=upload_dir,
                custom_metadata=custom_metadata,
            )

    @exception_handler
    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def predict_bytes(
        self,
        filename: str,
        raw_image: bytes,
        tags: str | Tag | list[Tag | str] | None = None,
        source: str | DataSource | None = None,
        metadata: dict | None = None,
        monitor: bool = True,
        upload_dir: str | None = None,
        custom_metadata: dict | None = None,
    ) -> dict:
        """Run a prediction on our Serving platform with bytes of an image

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            filename = "frame.png"
            with open(filename, 'rb') as img:
                img_bytes = img.read()
            deployment.predict_bytes(filename, img_bytes, tags=["tag1", "tag2"], source="camera-1")
            ```

        Arguments:
            filename (str): filename of the image.
            raw_image (bytes): bytes of the image to predict.
            tags (str, (Tag), list[str | Tag], optional): a list of tag to add to the data that will be created on the platform.
            source (str or DataSource, optional): a source to attach to the data that will be created on the platform.
            metadata (dict, optional): metadata to attach to the data that will be created on the platform.
            monitor (bool, optional): if True, will send prediction on Picsellia and our monitoring service. Defaults to True.
            upload_dir (str, optional): This parameter can only be used with private object-storages. Specify this parameter to prefix the object name of the data. Filename will still contain a generated uuid4
            custom_metadata (dict, optional): custom_metadata to attach to the data that will be created on the platform.

        Returns:
            A (dict) with information of the prediction
        """
        payload = self._prepare_payload_for_prediction(
            tags, source, monitor, upload_dir
        )
        files = {"media": (filename, raw_image)}

        # as a dict cannot be sent into a multipart upload, we need to stringify it for the serving
        if metadata:
            payload["metadata"] = json.dumps(metadata)
        if custom_metadata:
            payload["custom_metadata"] = json.dumps(custom_metadata)

        response = self.serving_connexion.post(
            path=f"/api/deployments/{self.id}/predict",
            data=payload,
            files=files,
        )
        if response.status_code > 500:
            raise MonitoringConnectionError()

        if response.status_code != 200:  # pragma: no cover
            raise PredictionError(f"Could not predict because {response.text}")

        return response.json()

    @exception_handler
    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def predict_cloud_image(
        self,
        object_name: str,
        tags: str | Tag | list[Tag | str] | None = None,
        source: str | DataSource | None = None,
        metadata: dict | None = None,
        monitor: bool = True,
        custom_metadata: dict | None = None,
    ) -> dict:
        """Run a prediction on our Serving platform, using object_name of a cloud object stored in your object storage.
        Your image MUST be stored in the storage used in the datalake linked to this deployment (the target datalake)
        If your image is already a data in your target datalake, it MUST NOT have been processed by this deployment,
            also, in this case, given source and metadata won't be used.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            object_name = "directory/s3/object-name.jpeg"
            deployment.predict_cloud_image(object_name, tags=["tag1", "tag2"], source="camera-1")
            ```

        Arguments:
            object_name (str): object name of the cloud image.
            tags (str, (Tag), list[str | Tag], optional): a list of tag to add to the data that will be created on the platform.
            source (str or DataSource, optional): a source to attach to the data that will be created on the platform.
            metadata (dict, optional): metadata to attach to the data that will be created on the platform.
            monitor (bool, optional): if True, will send prediction on Picsellia and our monitoring service. Defaults to True.
            custom_metadata (dict, optional): custom_metadata to attach to the data that will be created on the platform.

        Returns:
            A (dict) with information of the prediction
        """
        payload = self._prepare_payload_for_prediction(
            tags, source, monitor, upload_dir=None
        )
        if metadata:
            payload["metadata"] = metadata

        if custom_metadata:
            payload["custom_metadata"] = custom_metadata

        target_datalake_connector = self._fetch_target_datalake_connector()
        payload["cloud_image"] = {
            "connector_id": target_datalake_connector.id,
            "client_type": target_datalake_connector.client_type,
            "bucket_name": target_datalake_connector.bucket_name,
            "object_name": object_name,
        }

        response = self.serving_connexion.post(
            path=f"/api/v2/deployments/{self.id}/predict",
            data=orjson.dumps(payload),
        )
        if response.status_code > 500:
            raise MonitoringConnectionError()

        if response.status_code != 200:  # pragma: no cover
            raise PredictionError(f"Could not predict because {response.text}")

        return response.json()

    @exception_handler
    @beartype
    def predict_data(
        self,
        data: Data,
        tags: str | Tag | list[Tag | str] | None = None,
        monitor: bool = True,
    ) -> dict:
        """Run a prediction on our Serving platform, using data.
        Your data must already be stored in the datalake used by the deployment (the target datalake).
        If there already is a prediction in this deployment linked to this data, it will be dismissed.

        Specified tags will be added to the ones already existing on the Data.

        Examples:
            ```python
            datalake = client.get_datalake(name="target-datalake")
            data = datalake.list_data(limit=1)[0]
            deployment = client.get_deployment(name="awesome-deploy")
            deployment.predict_data(data, tags=["tag1", "tag2"], monitor=False)
            ```

        Arguments:
            data (Data): object that you want to predict on.
            tags (str, (Tag), list[str | Tag], optional): a list of tag to add to the data
            monitor (bool, optional): if True, will send prediction on Picsellia and our monitoring service. Defaults to True.

        Returns:
            A (dict) with information of the prediction
        """
        return self.predict_cloud_image(data.object_name, tags=tags, monitor=monitor)

    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def predict_shadow(
        self, predicted_asset: PredictedAsset, monitor: bool = True
    ) -> dict:
        """Add a shadow prediction on a predicted asset.
        It will call our Serving platform, returning predictions coming from shadow model.
        If monitor is true, it will go to our monitoring service, then it will be added on the platform.

        Examples:
            ```python
            deployment = client.get_deployment(name="awesome-deploy")
            predicted_asset = deployment.list_predicted_assets(limit=1)[0]
            deployment.predict_shadow(predicted_asset, monitor=False)
            ```

        Arguments:
            predicted_asset (PredictedAsset): shadow model will predict on this asset.
            monitor (bool, optional): if True, will send prediction on Picsellia and our monitoring service. Defaults to True.

        Returns:
            A (dict) with prediction shapes
        """
        target_datalake_connector = self._fetch_target_datalake_connector()
        payload = {
            "prediction_id": predicted_asset.oracle_prediction_id,
            "monitor": monitor,
            "cloud_image": {
                "connector_id": target_datalake_connector.id,
                "client_type": target_datalake_connector.client_type,
                "bucket_name": target_datalake_connector.bucket_name,
                "object_name": predicted_asset.object_name,
            },
        }
        response = self.serving_connexion.post(
            path=f"/api/v2/deployments/{self.id}/predict/shadow",
            data=orjson.dumps(payload),
        )
        if response.status_code > 500:
            raise MonitoringConnectionError

        if response.status_code != 200:  # pragma: no cover
            raise PredictionError(f"Could not predict shadow because {response.text}")

        return response.json()

    def _fetch_target_datalake_connector(
        self, force_refresh: bool = False
    ) -> TargetDatalakeConnectorSchema:
        if self._target_datalake_connector and not force_refresh:
            return self._target_datalake_connector

        r = self.connexion.get(f"/api/deployment/{self.id}/datalake/connector").json()
        self._target_datalake_connector = TargetDatalakeConnectorSchema(**r)
        return self._target_datalake_connector

    @staticmethod
    def _prepare_payload_for_prediction(
        tags: str | Tag | list[Tag | str] | None,
        source: str | DataSource | None,
        monitor: bool,
        upload_dir: str | None,
    ) -> dict[str, Any]:
        sent_tags = []
        if tags:
            if isinstance(tags, str) or isinstance(tags, Tag):
                tags = [tags]

            for tag in tags:
                if isinstance(tag, Tag):
                    sent_tags.append(tag.name)
                else:
                    sent_tags.append(tag)

        if isinstance(source, DataSource):
            source = source.name

        payload: dict[str, Any] = {"tags": sent_tags, "monitor": monitor}
        if source:
            payload["source"] = source
        if upload_dir:
            payload["upload_dir"] = upload_dir
        return payload

    @exception_handler
    @beartype
    def setup_feedback_loop(
        self, dataset_version: DatasetVersion | None = None
    ) -> None:
        """Set up the Feedback Loop for a Deployment.
        You can specify one Dataset Version to attach to it or use the
        attach_dataset_to_feedback_loop() afterward,so you can add multiple ones.
        This is a great option to increase your training set with quality data.

        Examples:
            ```python
            dataset_version = client.get_dataset("my-dataset").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_feedback_loop(dataset_version)
            ```

        Arguments:
            dataset_version (DatasetVersion, optional): This parameter is deprecated. Use attach_dataset_to_feedback_loop() instead.
        """
        self.connexion.post(
            f"/api/deployment/{self.id}/pipeline/fl/setup", data=orjson.dumps({})
        )
        logger.info(f"Feedback loop set for {self}")

        if dataset_version:
            self.attach_dataset_version_to_feedback_loop(dataset_version)
            logger.warning(
                "`dataset_version` parameter will be deprecated in future versions. "
                "Please call the attach_dataset_version_to_feedback_loop() after setup with the desired "
                "Dataset Version instead"
            )

    @exception_handler
    @beartype
    def attach_dataset_version_to_feedback_loop(
        self, dataset_version: DatasetVersion
    ) -> None:
        """Attach a Dataset Version to a previously configured feedback-loop.

        Examples:
            ```python
            dataset_versions = client.get_dataset("my-dataset").list_versions()
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_feedback_loop()
            for dataset_version in dataset_versions:
                deployment.attach_dataset_version_to_feedback_loop(dataset_version)
            ```

        Arguments:
            dataset_version (DatasetVersion): a (DatasetVersion) to attach
        """
        payload = {
            "dataset_version_id": dataset_version.id,
        }
        self.connexion.post(
            f"/api/deployment/{self.id}/fl/datasets",
            data=orjson.dumps(payload),
        )

    @exception_handler
    @beartype
    def detach_dataset_version_from_feedback_loop(
        self, dataset_version: DatasetVersion
    ) -> None:
        """Detach a Dataset Version from a previously configured feedback-loop.

        Examples:
            ```python
            dataset_versions = client.get_dataset("my-dataset").list_versions()
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_feedback_loop()
            for dataset_version in dataset_versions:
                deployment.attach_dataset_version_to_feedback_loop(dataset_version)
            deployment.detach_dataset_version_from_feedback_loop(dataset_versions[0])
            ```

        Arguments:
            dataset_version (DatasetVersion): a (DatasetVersion) to detach
        """

        payload = {"ids": [dataset_version.id]}
        self.connexion.delete(
            f"/api/deployment/{self.id}/fl/datasets",
            data=orjson.dumps(payload),
        )

    @exception_handler
    @beartype
    def list_feedback_loop_datasets(self) -> list[DatasetVersion]:
        """List the Dataset Versions attached to the feedback-loop

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            dataset_versions = deployment.list_feedback_loop_datasets()
            ```
        Returns:
            A list of (DatasetVersion)
        """
        r = self.connexion.get(f"/api/deployment/{self.id}/fl/datasets").json()
        return [
            DatasetVersion(self.connexion, item["dataset_version"])
            for item in r["items"]
        ]

    @exception_handler
    @beartype
    def toggle_feedback_loop(self, active: bool) -> None:
        """Toggle feedback loop for this deployment

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.toggle_feedback_loop(
                True
            )
            ```
        Arguments:
            active (bool): (des)activate feedback loop
        """
        payload = {"active": active}
        self.connexion.put(
            f"/api/deployment/{self.id}/pipeline/fl",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Feedback loop for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    def set_training_data(self, dataset_version: DatasetVersion) -> None:
        """This will give the training data reference to the deployment,
         so we can compute metrics based on this training data distribution in our Monitoring service

        Examples:
            ```python
            dataset_version = client.get_dataset("my-dataset").get_version("latest")
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.set_training_data(dataset_version)
            ```
        Arguments:
            dataset_version (DatasetVersion): a (DatasetVersion)
        """
        payload = {
            "dataset_version_id": dataset_version.id,
        }
        self.connexion.post(
            f"/api/deployment/{self.id}/pipeline/td/setup",
            data=orjson.dumps(payload),
        )
        logger.info(f"Training Data set for {self} from {dataset_version}")

    @exception_handler
    @beartype
    def check_training_data_metrics_status(self) -> str:
        """Refresh the status of the metrics compute over the training data distribution.
        Set up can take some time, so you can check current state with this method.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.check_training_data_metrics_status()
            ```
        Returns:
            A string with the status of the metrics compute over the training data distribution
        """
        r = self.connexion.get(f"/api/deployment/{self.id}/pipeline/td/check").json()
        status = r["status"]
        logger.info(f"Training Data status is {status}")
        return status

    @exception_handler
    @beartype
    def disable_training_data_reference(self) -> None:
        """Disable the reference to the training data in this Deployment.
        This means that you will not be able to see supervised metrics from the dashboard anymore.

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.disable_training_data_reference()
            ```
        """
        self.connexion.put(f"/api/deployment/{self.id}/pipeline/td/disable")
        logger.info(f"Training Data for {self} is disabled.")

    @exception_handler
    @beartype
    def setup_continuous_training(
        self,
        project: Project,
        dataset_version: DatasetVersion | None = None,
        model_version: ModelVersion | None = None,
        trigger: str | ContinuousTrainingTrigger = None,
        threshold: int | None = None,
        experiment_parameters: dict | None = None,
        scan_config: dict | None = None,
    ) -> None:
        """Initialize and activate the continuous training features of picsellia. 🥑
           A Training will be triggered using the attached dataset versions
           whenever your Deployment pipeline hit the trigger.
           You can call attach_dataset_version_to_continuous_training() method afterward.
           You can launch a continuous training via Experiment with parameter `experiment_parameters`
           You cannot launch a continuous training via Scan at the moment

        Examples:
            We want to set up a continuous training pipeline that will be trigger
            every 150 new predictions reviewed by your team.
            We will use the same training parameters as those used when building the first model.

            ```python
            deployment = client.get_deployment("awesome-deploy")
            project = client.get_project(name="my-project")
            dataset_version = project.get_dataset(name="my-dataset").get_version("latest")
            model_version = client.get_model(name="my-model").get_version(0)
            experiment = model_version.get_source_experiment()
            experiment_parameters = experiment.get_log('parameters')
            deployment.setup_continuous_training(
                project, threshold=150, experiment_parameters=experiment_parameters
            )
            deployment.attach_continuous
            ```
        Arguments:
            project (Project): The project that will host your pipeline.
            dataset_version (DatasetVersion, deprecated): This parameter is deprecated and is not used anymore.
            model_version (ModelVersion, deprecated): This parameter is deprecated and is not used anymore.
            threshold (int): Number of images that need to be review to trigger the training.
            trigger (ContinuousTrainingTrigger): Type of trigger to use when there is enough reviews.
            experiment_parameters (dict, optional):  Training parameters. Defaults to None.
        """
        payload: dict[str, Any] = {"project_id": project.id}

        if model_version:
            logger.warning(
                "`model_version` parameter is no longer used. "
                "Continuous Training will be configured with deployment's model version"
            )

        if dataset_version:
            logger.warning(
                "`dataset_version` parameter will be deprecated in future versions. "
                "Please call the attach_dataset_version_to_continuous_training() after setup with the desired "
                "Dataset Versions instead"
            )
        if trigger is not None and threshold is not None:
            payload["trigger"] = ContinuousTrainingTrigger.validate(trigger)
            payload["threshold"] = threshold

        if scan_config:
            logger.warning("`scan_config` parameter is no longer used.")

        if not experiment_parameters:
            raise BadConfigurationContinuousTrainingError(
                "You need to give `experiment_parameters`"
            )

        payload["training_type"] = ContinuousTrainingType.EXPERIMENT
        payload["experiment_parameters"] = experiment_parameters

        self.connexion.post(
            f"/api/deployment/{self.id}/pipeline/ct",
            data=orjson.dumps(payload),
        )
        logger.info(f"Continuous training setup for {self}\n")

    @exception_handler
    @beartype
    def attach_dataset_version_to_continuous_training(
        self, alias: str, dataset_version: DatasetVersion
    ):
        """Attach a Dataset Version to a previously configured continuous training.

        Examples:
            ```python
            dataset_versions = client.get_dataset("my-dataset").list_versions()
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_continuous_training(...)
            aliases = ["train", "test", "eval"]
            for i, dataset_version in enumerate(dataset_versions):
                deployment.attach_dataset_version_to_continuous_training(aliases[i], dataset_version)
            ```
        Arguments:
            alias (str): Alias of attached dataset
            dataset_version (DatasetVersion): A dataset version to attach to the Continuous Training.
        """
        payload = {"dataset_version_id": dataset_version.id, "name": alias}
        self.connexion.post(
            f"/api/deployment/{self.id}/ct/datasets",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"{dataset_version} attached to Continuous training of {self} with alias {alias}"
        )

    @exception_handler
    @beartype
    def detach_dataset_version_from_continuous_training(
        self, dataset_version: DatasetVersion
    ) -> None:
        """Detach a Dataset Versions to a previously configured continuous training.

        Examples:
            ```python
            dataset_versions = client.get_dataset("my-dataset").list_versions()
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_continuous_training()
            for dataset_version in dataset_versions:
                deployment.attach_dataset_version_to_continuous_training(dataset_version)
            deployment.detach_dataset_version_from_continuous_training(dataset_versions[0])
            ```

        Arguments:
            dataset_version (DatasetVersion): a (DatasetVersion) to detach from Continuous Training settings
        """

        payload = {"ids": [dataset_version.id]}
        self.connexion.delete(
            f"/api/deployment/{self.id}/ct/datasets",
            data=orjson.dumps(payload),
        )

    @exception_handler
    @beartype
    def toggle_continuous_training(self, active: bool) -> None:
        """Toggle continuous training for this deployment

        Examples:
            ```python
            deployment = client.get_deployment("awesome-deploy")
            deployment.toggle_continuous_training(active=False)
            ```

        Arguments:
            active (bool): (des)activate continuous training
        """
        payload = {"active": active}
        self.connexion.put(
            f"/api/deployment/{self.id}/pipeline/ct",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Continuous training for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    def setup_continuous_deployment(
        self, policy: ContinuousDeploymentPolicy | str
    ) -> None:
        """Set up the continuous deployment for this pipeline

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.setup_continuous_deployment(ContinuousDeploymentPolicy.DEPLOY_MANUAL)
            ```
        Arguments:
            policy (ContinuousDeploymentPolicy): policy to use
        """
        payload = {"policy": ContinuousDeploymentPolicy.validate(policy)}
        self.connexion.post(
            f"/api/deployment/{self.id}/pipeline/cd",
            data=orjson.dumps(payload),
        )
        logger.info(f"Continuous deployment setup for {self} with policy {policy}\n")

    @exception_handler
    @beartype
    def toggle_continuous_deployment(self, active: bool) -> None:
        """Toggle continuous deployment for this deployment

        Examples:
            ```python
            deployment = client.get_deployment(
                name="awesome-deploy"
            )
            deployment.toggle_continuous_deployment(
                dataset
            )
            ```
        Arguments:
            active (bool): (des)activate continuous deployment
        """
        payload = {"active": active}
        self.connexion.put(
            f"/api/deployment/{self.id}/pipeline/cd",
            data=orjson.dumps(payload),
        )
        logger.info(
            f"Continuous deployment for {self} is now {'active' if active else 'deactivated'}"
        )

    @exception_handler
    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def get_stats(
        self,
        service: ServiceMetrics,
        model_version: ModelVersion | None = None,
        from_timestamp: float | None = None,
        to_timestamp: float | None = None,
        since: int | None = None,
        includes: list[str] | None = None,
        excludes: list[str] | None = None,
        tags: list[str] | None = None,
    ) -> dict:
        """Retrieve stats of this deployment stored in Picsellia environment.

        Mandatory param is "service" an enum of type ServiceMetrics. Values possibles are :
            PREDICTIONS_OUTLYING_SCORE
            PREDICTIONS_DATA
            REVIEWS_OBJECT_DETECTION_STATS
            REVIEWS_CLASSIFICATION_STATS
            REVIEWS_LABEL_DISTRIBUTION_STATS

            AGGREGATED_LABEL_DISTRIBUTION
            AGGREGATED_OBJECT_DETECTION_STATS
            AGGREGATED_PREDICTIONS_DATA
            AGGREGATED_DRIFTING_PREDICTIONS

        For aggregation, computation may not have been done by the past.
        You will need to force computation of these aggregations and retrieve them again.


        Examples:
            ```python
            my_deployment.get_stats(ServiceMetrics.PREDICTIONS_DATA)
            my_deployment.get_stats(ServiceMetrics.AGGREGATED_DRIFTING_PREDICTIONS, since=3600)
            my_deployment.get_stats(ServiceMetrics.AGGREGATED_LABEL_DISTRIBUTION, model_version=my_model)
            ```

        Arguments:
            service (str): service queried
            model_version (ModelVersion, optional): Model that shall be used when retrieving data.
                Defaults to None.
            from_timestamp (float, optional): System will only retrieve prediction data after this timestamp.
                Defaults to None.
            to_timestamp (float, optional): System will only retrieve prediction data before this timestamp.
                Defaults to None.
            since (int, optional): System will only retrieve prediction data that are in the last seconds.
                Defaults to None.
            includes (list[str], optional): Research will include these ids and excludes others.
                Defaults to None.
            excludes (list[str], optional): Research will exclude these ids.
                Defaults to None.
            tags (list[str], optional): Research will be done filtering by tags.
                Defaults to None.

        Returns:
            A dict with queried statistics about the service you asked
        """
        query_filter = self._build_filter(
            service=service.service,
            model_version=model_version,
            from_timestamp=from_timestamp,
            to_timestamp=to_timestamp,
            since=since,
            includes=includes,
            excludes=excludes,
            tags=tags,
        )

        if service.is_aggregation:
            response = self.oracle_connexion.get(
                path=f"/api/deployment/{self.id}/stats", params=query_filter
            )
            if response.status_code > 500:
                raise MonitoringConnectionError()
            content = response.json()
            if "infos" in content and "info" in content["infos"]:
                logger.info("This computation is outdated or has never been done.")
                logger.info(
                    "You can compute it again by calling launch_computation with exactly the same params."
                )
            return content
        else:
            response = self.oracle_connexion.get(
                path=f"/api/deployment/{self.id}/predictions/stats",
                params=query_filter,
            )
            if response.status_code > 500:
                raise MonitoringConnectionError()
            return response.json()

    @staticmethod
    def _build_filter(
        service: str,
        model_version: ModelVersion | None = None,
        from_timestamp: float | None = None,
        to_timestamp: float | None = None,
        since: int | None = None,
        includes: list[str] | None = None,
        excludes: list[str] | None = None,
        tags: list[str] | None = None,
    ) -> dict[str, Any]:
        query_filter: dict[str, Any] = {"service": service}

        if model_version is not None:
            query_filter["model_id"] = model_version.id

        if from_timestamp is not None:
            query_filter["from_timestamp"] = from_timestamp

        if to_timestamp is not None:
            query_filter["to_timestamp"] = to_timestamp

        if since is not None:
            query_filter["since"] = since

        if includes is not None:
            query_filter["includes"] = includes

        if excludes is not None:
            query_filter["excludes"] = excludes

        if tags is not None:
            query_filter["tags"] = tags

        return query_filter

    @exception_handler
    @beartype
    def monitor(
        self,
        image_path: str | Path,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        source: str | None = None,
        tags: list[str] | None = None,
        timestamp: float | None = None,
        model_version: ModelVersion | None = None,
        shadow_model_version: ModelVersion | None = None,
        shadow_latency: float | None = None,
        shadow_raw_predictions: PredictionFormat | None = None,
        shadow_prediction: PredictionFormat | None = None,
        content_type: SupportedContentType | str | None = None,
        metadata: dict | None = None,
        upload_dir: str | None = None,
        custom_metadata: dict | None = None,
        sample: bool | None = None,
    ) -> dict:
        """Send a prediction for this deployment on our monitoring service.

        :warning: Signature of this method has been recently changed and can break some methods :
        - model_version and shadow_model_version are not used anymore : system will use what's currently being monitored in this deployment
        - shadow_raw_predictions has been renamed to shadow_prediction

        Arguments:
            image_path (str or Path): image path
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (list[int]): list of classes
                    detection_scores (list[float]): list of scores of predictions
                DetectionPredictionFormat and SegmentationPredictionFormat:
                    detection_boxes (list[list[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks  list[list[list[int]]]): list of polygon shapes, each element corresponds to a distinct shape and is a list of the coordinates
                                                            one polygon will look like [[x1, y1], [x2, y2], [x3, y3], [x4, y4], ...]
                LinePredictionFormat:
                    lines (list[list[int]]): list of line shapes, each element corresponds to a distinct shape and is a list of the coordinates
                                                            one line will look like [[x1, y1], [x2, y2], [x3, y3], [x4, y4], ...]
                PointPredictionFormat:
                    points (list[list[int]]): list of point shapes, each element corresponds to a distinct point and is the coordinates
                                                            a point will look like [x, y]

            source (str, optional): (Data) will have this source in Picsellia. Defaults to None.
            tags (list[str], optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.
            content_type (str, optional): if given, we won't try to infer content type with mimetype library
            metadata (dict, optional): (Data) will have this metadata in Picsellia. Defaults to None.
            upload_dir (str, optional): This parameter can only be used with private object-storages. Specify this parameter to prefix the object name of the data. Filename will still contain a generated uuid4
            custom_metadata (dict, optional): (Data) will have this custom_metadata in Picsellia. Defaults to None.
            sample (bool, optional): control sample policy by using this field.
                    If True, the Data and the Prediction will be stored under the Deployment (no matter the sample_rate).
                    If False, the Data and the Prediction won't be stored, monitoring metrics WILL still be computed.
                    If unset, we will store the Data and Prediction according to the sample_rate (if existing).
        Returns:
            a dict of data returned by our monitoring service
        """
        if model_version:  # pragma: no cover
            logger.warning(
                "'model_version' will soon be removed. It is not used anymore"
            )

        if shadow_model_version:  # pragma: no cover
            logger.warning(
                "'shadow_model_version' will soon be removed. It is not used anymore"
            )

        if shadow_raw_predictions and not shadow_prediction:  # pragma: no cover
            logger.warning(
                "'shadow_raw_predictions' parameter will soon be removed. Please use 'shadow_prediction'"
            )
            shadow_prediction = shadow_raw_predictions

        if not content_type:
            content_type = mimetypes.guess_type(image_path, strict=False)[0]
            if content_type is None:  # pragma: no cover
                raise ContentTypeUnknown(
                    f"Content type of {image_path} could not be inferred"
                )

        # Open image, encode it into base64, read filename and content type.
        with open(image_path, "rb") as img_file:
            raw_image = img_file.read()
            filename = Path(image_path).name

        return self.monitor_bytes(
            raw_image,
            content_type,
            filename,
            latency,
            height,
            width,
            prediction,
            source,
            tags,
            timestamp,
            shadow_latency,
            shadow_prediction,
            metadata,
            upload_dir,
            custom_metadata,
            sample,
        )

    @exception_handler
    @beartype
    def monitor_bytes(
        self,
        raw_image: bytes,
        content_type: SupportedContentType | str,
        filename: str,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        source: str | None = None,
        tags: list[str] | None = None,
        timestamp: float | None = None,
        shadow_latency: float | None = None,
        shadow_prediction: PredictionFormat | None = None,
        metadata: dict | None = None,
        upload_dir: str | None = None,
        custom_metadata: dict | None = None,
        sample: bool | None = None,
    ) -> dict:
        """Send a prediction for this deployment on our monitoring service.
        You can use this method instead of monitor() if you have a bytes image and not an image file.
        We will convert it into base 64 as utf8 string and send it to the monitoring service.

        Arguments:
            raw_image (bytes): raw image in bytes
            content_type (SupportedContentType or str): content type of image, only 'image/jpeg' or 'image/png' currently supported
            filename (str): filename of image
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (list[int]): list of classes
                    detection_scores (list[float]): list of scores of predictions
                DetectionPredictionFormat:
                    detection_boxes (list[list[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks (list[list[int]]): list of polygons of your shapes. each polygon is a list of points with coordinates flattened
                                                            [x1, y1, x2, y2, x3, y3, x4, y4, ..]

            source (str, optional): source that can give some metadata to your prediction. Defaults to None.
            tags (list[str], optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.
            metadata (dict, optional): (Data) will have this metadata in Picsellia. Defaults to None.
            upload_dir (str, optional): This parameter can only be used with private object-storages. Specify this parameter to prefix the object name of the data. Filename will still contain a generated uuid4
            custom_metadata (dict, optional): (Data) will have this custom_metadata in Picsellia. Defaults to None.
            sample (bool, optional): control sample policy by using this field.
                    If True, the Data and the Prediction will be stored under the Deployment (no matter the sample_rate).
                    If False, the Data and the Prediction won't be stored, monitoring metrics WILL still be computed.
                    If unset, we will store the Data and Prediction according to the sample_rate (if existing).
        Returns:
            a dict of data returned by our monitoring service
        """
        payload = self._prepare_payload_monitor(
            filename,
            latency,
            height,
            width,
            prediction,
            content_type,
            source,
            tags,
            timestamp,
            shadow_latency,
            shadow_prediction,
            metadata,
            upload_dir,
            custom_metadata,
            sample,
        )
        # Convert bytes into a base 64 string
        payload["image"] = base64.b64encode(raw_image).decode("utf-8")
        return self._request_monitor(payload)

    @exception_handler
    @beartype
    def monitor_cloud_image(
        self,
        object_name: str,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        content_type: SupportedContentType | str,
        source: str | None = None,
        tags: list[str] | None = None,
        timestamp: float | None = None,
        shadow_latency: float | None = None,
        shadow_prediction: PredictionFormat | None = None,
        metadata: dict | None = None,
        custom_metadata: dict | None = None,
        sample: bool | None = None,
    ) -> dict:
        """Monitor an image on our monitoring platform, using object_name of a cloud object stored in your object storage.

        Arguments:
            object_name (str): object name of the cloud image.
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (list[int]): list of classes
                    detection_scores (list[float]): list of scores of predictions
                DetectionPredictionFormat and SegmentationPredictionFormat:
                    detection_boxes (list[list[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks (list[list[int]]): list of polygons of your shapes. each polygon is a list of points with coordinates flattened
                                                            [x1, y1, x2, y2, x3, y3, x4, y4, ..]

            source (str, optional): (Data) will have this source in Picsellia. Defaults to None.
            tags (list[str], optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.
            content_type (str, optional): if given, we won't try to infer content type with mimetype library
            metadata (dict, optional): (Data) will have this metadata in Picsellia. Defaults to None.
            custom_metadata (dict, optional): (Data) will have this custom_metadata in Picsellia. Defaults to None.
            sample (bool, optional): control sample policy by using this field.
                    If True, the Data and the Prediction will be stored under the Deployment (no matter the sample_rate).
                    If False, the Data and the Prediction won't be stored, monitoring metrics WILL still be computed.
                    If unset, we will store the Data and Prediction according to the sample_rate (if existing).
        Returns:
            a dict of data returned by our monitoring service
        """
        filename = os.path.basename(object_name)
        payload = self._prepare_payload_monitor(
            filename,
            latency,
            height,
            width,
            prediction,
            content_type,
            source,
            tags,
            timestamp,
            shadow_latency,
            shadow_prediction,
            metadata,
            upload_dir=None,
            custom_metadata=custom_metadata,
            sample=sample,
        )

        target_datalake_connector = self._fetch_target_datalake_connector()
        payload["cloud_image"] = {
            "connector_id": target_datalake_connector.id,
            "client_type": target_datalake_connector.client_type,
            "bucket_name": target_datalake_connector.bucket_name,
            "object_name": object_name,
        }

        return self._request_monitor(payload)

    @exception_handler
    @beartype
    def monitor_data(
        self,
        data: Data,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        tags: list[str] | None = None,
        timestamp: float | None = None,
        shadow_latency: float | None = None,
        shadow_prediction: PredictionFormat | None = None,
        sample: bool | None = None,
    ) -> dict:
        """Monitor an image on our monitoring platform
        Your data must already be stored in the datalake used by the deployment (the target datalake).
        If there already is a prediction in this deployment linked to this data, it will be dismissed.

        Arguments:
            data (Data): data to monitor
            latency (float): latency used by model to compute your prediction
            height (int): height of image
            width (int): width of image
            prediction (PredictionFormat): data of your prediction, can be a Classification, a Segmentation or an ObjectDetection Format.
                DetectionPredictionFormat, ClassificationPredictionFormat and SegmentationPredictionFormat:
                    detection_classes (list[int]): list of classes
                    detection_scores (list[float]): list of scores of predictions
                DetectionPredictionFormat and SegmentationPredictionFormat:
                    detection_boxes (list[list[int]]): list of bboxes representing rectangles of your shapes. bboxes are formatted as
                                                            [top, left, bottom, right]
                SegmentationPredictionFormat:
                    detection_masks (list[list[int]]): list of polygons of your shapes. each polygon is a list of points with coordinates flattened
                                                            [x1, y1, x2, y2, x3, y3, x4, y4, ..]

            tags (list[str], optional): tags that can give some metadata to your prediction. Defaults to None.
            timestamp (float, optional): timestamp of your prediction. Defaults to timestamp of monitoring service on reception.
            shadow_latency (float, optional): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat, optional): data of your prediction made by shadow model.
            sample (bool, optional): control sample policy by using this field.
                    If True, the Data and the Prediction will be stored under the Deployment (no matter the sample_rate).
                    If False, the Data and the Prediction won't be stored, monitoring metrics WILL still be computed.
                    If unset, we will store the Data and Prediction according to the sample_rate (if existing).

        Returns:
            a dict of data returned by our monitoring service
        """
        return self.monitor_cloud_image(
            data.object_name,
            latency,
            height,
            width,
            prediction,
            data.content_type,
            source=None,
            tags=tags,
            timestamp=timestamp,
            shadow_latency=shadow_latency,
            shadow_prediction=shadow_prediction,
            metadata=None,
            custom_metadata=None,
            sample=sample,
        )

    def _prepare_payload_monitor(  # noqa: C901
        self,
        filename: str,
        latency: float,
        height: int,
        width: int,
        prediction: PredictionFormat,
        content_type: SupportedContentType | str,
        source: str | None,
        tags: list[str] | None,
        timestamp: float | None,
        shadow_latency: float | None,
        shadow_prediction: PredictionFormat | None,
        metadata: dict | None,
        upload_dir: str | None,
        custom_metadata: dict | None,
        sample: bool | None,
    ):
        if prediction.model_type != self.type:
            raise BadRequestError(
                f"Prediction shape of this type {prediction.model_type} cannot be used with this model {self.type}"
            )

        try:
            content_type = SupportedContentType.validate(content_type)
        except TypeError:
            raise ContentTypeUnknown(
                f"Content type {content_type} is not supported : {SupportedContentType.values()}"
            )

        payload = {
            "filename": filename,
            "content_type": content_type.value,
            "height": height,
            "width": width,
            "raw_predictions": prediction.model_dump(),
            "latency": latency,
        }

        if source is not None:
            payload["source"] = source

        if tags is not None:
            payload["tags"] = tags

        if metadata is not None:
            payload["metadata"] = metadata

        if custom_metadata is not None:
            payload["custom_metadata"] = custom_metadata

        if timestamp is not None:
            payload["timestamp"] = timestamp

        if upload_dir is not None:
            payload["upload_dir"] = upload_dir

        if shadow_prediction is not None:
            if shadow_latency is None:
                raise BadRequestError(
                    "Shadow latency and shadow raw predictions shall be defined if you want to push a shadow result"
                )
            if shadow_prediction.model_type != self.type:
                raise BadRequestError(
                    f"Prediction shape of this type {prediction.model_type} cannot be used with this model {self.type}"
                )

            payload["shadow_latency"] = shadow_latency
            payload["shadow_raw_predictions"] = shadow_prediction.model_dump()

        if sample is not None:
            payload["sample"] = sample

        return payload

    @retry((requests.ConnectionError, MonitoringConnectionError))
    def _request_monitor(self, payload: dict) -> dict:
        response = self.oracle_connexion.post(
            path=f"/api/deployment/{self.id}/predictions",
            data=orjson.dumps(payload),
        )
        if response.status_code > 500:
            raise MonitoringConnectionError()

        if response.status_code == 409:
            error = response.json()["message"]
            raise MonitorError(f"This data has already been processed : {error}")

        if response.status_code != 201:  # pragma: no cover
            raise MonitorError(
                f"Our monitoring service could not handle your prediction: {response.status_code}.\n Error : {response.text}"
            )

        return response.json()

    @exception_handler
    @beartype
    def monitor_shadow(
        self,
        predicted_asset: PredictedAsset,
        shadow_latency: float,
        shadow_prediction: PredictionFormat,
    ) -> None:
        """Add a shadow prediction on an existing PredictedAsset.
        You can call monitor_shadow_from_oracle_prediction_id() if you only have oracle_prediction_id

        Arguments:
            predicted_asset (PredictedAsset): asset already processed on which to add shadow_prediction
            shadow_latency (float): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat): data of your prediction made by shadow model
        """
        return self.monitor_shadow_from_oracle_prediction_id(
            predicted_asset.oracle_prediction_id, shadow_latency, shadow_prediction
        )

    @exception_handler
    @beartype
    @retry((requests.ConnectionError, MonitoringConnectionError))
    def monitor_shadow_from_oracle_prediction_id(
        self,
        oracle_prediction_id: str | UUID,
        shadow_latency: float,
        shadow_prediction: PredictionFormat,
    ) -> None:
        """Add a shadow prediction on an existing PredictedAsset, from the oracle_prediction_id

        Arguments:
            oracle_prediction_id (str or UUID): oracle_prediction_id that was returned on monitor()
            shadow_latency (float): latency used by shadow model to compute prediction
            shadow_prediction (PredictionFormat): data of your prediction made by shadow model
        """
        payload = {
            "shadow_raw_predictions": shadow_prediction.model_dump(),
            "shadow_latency": shadow_latency,
        }
        response = self.oracle_connexion.post(
            path=f"/api/prediction/{oracle_prediction_id}/shadow",
            data=orjson.dumps(payload),
        )
        if response.status_code > 500:
            raise MonitoringConnectionError()

        if response.status_code == 409:
            error = response.json()["message"]
            raise MonitorError(
                f"This prediction has already been shadow processed : {error}"
            )

        if response.status_code != 201:  # pragma: no cover
            raise MonitorError(
                f"Our monitoring service could not handle your shadow prediction: {response.status_code}. Check {response.text}"
            )

    @exception_handler
    @beartype
    def find_predicted_asset(
        self,
        id: str | UUID | None = None,
        oracle_prediction_id: str | UUID | None = None,
        object_name: str | None = None,
        filename: str | None = None,
        data_id: str | UUID | None = None,
    ) -> PredictedAsset:
        """Find a PredictedAsset of this deployment.

        Examples:
            ```python
            oracle_prediction_id = deployment.monitor(path, latency, height, width, prediction_data)["id"]
            predicted_asset = deployment.find_predicted_asset(oracle_prediction_id=oracle_prediction_id)
            deployment.monitor_shadow(predicted_asset, shadow_latency, shadow_prediction_data)
            ```
         Arguments:
            id (UUID, optional): id of PredictedAsset to fetch. Defaults to None.
            oracle_prediction_id (UUID, optional): id of the prediction in our monitoring system. Defaults to None.
            filename (str, optional): filename of the data. Defaults to None.
            object_name (str, optional): object_name of the data. Defaults to None.
            data_id (UUID, optional): id of the data related to this PredictedAsset. Defaults to None.

        Raises:
            If no asset match the query, it will raise a NotFoundError.
            In some case, it can raise an InvalidQueryError,
                it might be because the platform stores 2 assets matching this query (for example if filename is duplicated)

        Returns:
            The (PredictedAsset) found
        """
        if (
            not id
            and not oracle_prediction_id
            and not object_name
            and not filename
            and not data_id
        ):
            raise AssertionError(
                "Please select at least one criteria to find a predicted asset"
            )

        params: dict[str, Any] = {}
        if id:
            params["id"] = id

        if oracle_prediction_id:
            params["oracle_prediction_id"] = oracle_prediction_id

        if object_name:
            params["object_name"] = object_name

        if filename:
            params["filename"] = filename

        if data_id:
            params["data_id"] = data_id

        r = self.connexion.get(
            f"/api/deployment/{self.id}/predictedassets/find", params=params
        ).json()
        return PredictedAsset(self.connexion, self.id, r)

    @exception_handler
    @beartype
    def list_predicted_assets(
        self,
        limit: int | None = None,
        offset: int | None = None,
        page_size: int | None = None,
        order_by: list[str] | None = None,
        q: str | None = None,
        assignment_status: str | None = None,
        assignment_step_id: UUID | None = None,
        assignment_user_id: str | UUID | None = None,
        custom_metadata: dict | None = None,
    ) -> MultiPredictedAsset:
        """List (PredictedAsset) of this (Deployment)

        Examples:
            ```python
            assets = deployment.list_predicted_assets()
            ```

        Arguments:
            limit (int, optional): the number of assets that will be retrieved
            offset (int, optional): from where to start accessing the assets. you will retrieve offset:offset+limit assets from the whole list.
            page_size (int, optional): deprecated.
            order_by (str, optional): what property to sort on. Defaults to descending created_at.
            q (str, optional): a query using the Picsellia Query Language. Defaults to None
            assignment_status (str, optional): only with campaigns. It's the desired status for the last assignments linked to your (PredictedAsset).
            assignment_step_id (UUID, optional): only with campaigns. It's the desired step of the assignment you want to retrieve.
            assignment_user_id (UUID, optional): only with Campaigns. It's the desired user assigned to the assignments you want to retrieve.
            custom_metadata: (dict, optional): filter based on the custom_metadata linked to the asset's Data. Defaults to None
        Returns:
            A (MultiAsset) object that wraps some (Asset) that you can manipulate.
        """
        if page_size:
            logger.warning("page_size is deprecated and not used anymore.")

        filters = PredictedAssetFilter.model_validate(
            {
                "limit": limit,
                "offset": offset,
                "order_by": order_by,
                "query": q,
                "assignment_status": assignment_status,
                "assignment_step_id": assignment_step_id,
                "assignment_user_id": assignment_user_id,
                "custom_metadata": custom_metadata,
            }
        )
        assets = PredictedAssetLister(self.connexion, self.id).list_items(filters)

        if len(assets) == 0:
            raise NoDataError("No predicted asset retrieved")

        return MultiPredictedAsset(self.connexion, self.id, assets)

    @exception_handler
    @beartype
    @deprecated(
        deprecated_in="6.24.0",
        removed_in="6.27.0",
        details="This method should not be called anymore. Instead use list_users()",
    )
    def list_workers(self) -> list[Worker]:
        """List all workers of this deployment

        Examples:
            ```python
            deployment.list_workers()
            ```

        Returns:
            List of (Worker) objects
        """
        return [
            Worker(
                self.connexion,
                {"id": item.id, "username": item.username, "user_id": item.id},
                WorkerType.DEPLOYMENT,
            )
            for item in self.list_users()
        ]

    @exception_handler
    @beartype
    def create_campaign(
        self,
        description: str | None = None,
        instructions_file_path: str | None = None,
        instructions_text: str | None = None,
        end_date: date | None = None,
        auto_add_new_assets: bool | None = False,
        auto_close_on_completion: bool | None = False,
    ) -> ReviewCampaign:
        """Create campaign on a deployment.

        Examples:
            ```python
            foo_deployment.create_campaign()
            ```
        Arguments:
            description (str, optional): Description of the campaign. Defaults to None.
            instructions_file_path (str, optional): Instructions file path. Defaults to None.
            instructions_text (str, optional): Instructions text. Defaults to None.
            end_date (date, optional): End date of the campaign. Defaults to None.
            auto_add_new_assets (bool, optional): If true, new assets of this deployment will be added as a task
                                                    in the campaign. Defaults to False.
            auto_close_on_completion (bool, optional): If true, campaign will be close when all tasks will be done.
                                                        Defaults to False.

        Returns:
            A (ReviewCampaign) object
        """
        payload = {
            "description": description,
            "instructions_text": instructions_text,
            "end_date": end_date,
            "auto_add_new_assets": auto_add_new_assets,
            "auto_close_on_completion": auto_close_on_completion,
        }

        if instructions_file_path:
            instructions_file_name = os.path.basename(instructions_file_path)
            object_name = self.connexion.generate_deployment_object_name(
                instructions_file_name,
                ObjectDataType.REVIEW_CAMPAIGN_FILE,
                deployment_id=self.id,
            )
            self.connexion.upload_file(object_name, instructions_file_path)
            payload["instructions_object_name"] = object_name

        r = self.connexion.post(
            f"/api/deployment/{self.id}/campaigns",
            data=orjson.dumps(payload),
        ).json()
        campaign = ReviewCampaign(self.connexion, r)
        self._review_campaign_id = campaign.id
        logger.info(f"{campaign} has been created to {self}")
        return campaign

    @exception_handler
    @beartype
    def get_campaign(self) -> ReviewCampaign:
        """Get campaign of a dataset version.

        Examples:
            ```python
            foo_deployment.get_campaign()
            ```

        Returns:
            A (ReviewCampaign) object
        """
        if not self._review_campaign_id:
            self.sync()

        if not self._review_campaign_id:
            raise NoDataError("There is no review campaign defined in this deployment")

        try:
            r = self.connexion.get(
                f"/api/campaigns/review/{self._review_campaign_id}"
            ).json()
        except ResourceNotFoundError:
            self._review_campaign_id = None
            raise

        return ReviewCampaign(self.connexion, r)



import logging
from uuid import UUID

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.label_group import LabelGroup
from picsellia.types.schemas import LabelSchema

logger = logging.getLogger("picsellia")


class Label(Dao):
    def __init__(
        self, connexion: Connection, dataset_version_id: UUID, data: dict
    ) -> None:
        Dao.__init__(self, connexion, data)
        self._dataset_version_id = dataset_version_id

    @property
    def name(self) -> str:
        """Name of this (Label)"""
        return self._name

    @property
    def group_id(self) -> UUID | None:
        """Group ID of this (Label)"""
        return self._group_id

    def __str__(self):
        return f"{Colors.GREEN}Label '{self.name}'{Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/label/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> LabelSchema:
        schema = LabelSchema(**data)
        self._name = schema.name
        self._group_id = schema.group_id
        return schema

    @exception_handler
    @beartype
    def update(self, name: str = None, group_id: UUID = None) -> None:
        """Update this (Label) with a new name or with a new group.

        Examples:
            ```python
            a_label.update(name="new name")
            ```

        Arguments:
            name (str, optional): New name of this (Label)
            group_id (UUID, optional): New group of this (Label)
        """
        payload = {}
        if name:
            payload["name"] = name
        if group_id:
            payload["group_id"] = group_id
        r = self.connexion.patch(
            f"/api/label/{self.id}", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this (Label) from the platform.
        All annotations shape with this label will be deleted!
        This is a very dangerous move.

        :warning: **DANGER ZONE**: Be very careful here!

        Examples:
            ```python
            this_label.delete()
            ```
        """
        self.connexion.delete(f"/api/label/{self.id}")
        logger.info(f"{self} deleted.")

    @exception_handler
    @beartype
    def get_group(self) -> LabelGroup | None:
        """Get (LabelGroup) parent of this (Label). If no (LabelGroup) is set, return None.

        Examples:
            ```python
            group = label_1.get_group()
            label_2.set_group(group)
            ```

        Returns:
            a (LabelGroup) if there is a parent, else None
        """
        if not self._group_id:
            return None

        r = self.connexion.get(f"/api/label/group/{self._group_id}").json()
        return LabelGroup(self.connexion, self._dataset_version_id, r)

    @exception_handler
    @beartype
    def set_group(self, group: LabelGroup) -> None:
        """Update this (Label) with a new group.

        Examples:
            ```python
            a_label.update(name="new name")
            ```

        Arguments:
            group (LabelGroup): new group of this (Label)
        """
        if group.dataset_version_id != self._dataset_version_id:
            raise ValueError("Cannot set a group from a different DatasetVersion")

        return self.update(group_id=group.id)

    @exception_handler
    @beartype
    def get_skeleton(self) -> dict:
        """Get the skeleton of this (Label).
        Returns:
            a dict with key edges and vertices
        """
        return self.connexion.get(f"/api/label/{self.id}/skeleton").json()

    @exception_handler
    @beartype
    def set_skeleton(self, vertices: list[str], edges: list[list[int]]) -> None:
        """Set the skeleton of this (Label).

        Arguments:
            vertices (list[str]): names of this skeleton vertices
            edges (list of 2 ints): edges are links between vertices with indexes starting at 1!
        """
        payload = {"vertices": vertices, "edges": edges}
        self.connexion.post(
            f"/api/label/{self.id}/skeleton", data=orjson.dumps(payload)
        )
        logger.info(f"{self}' skeleton updated")



import logging

from beartype import beartype

from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.schemas import UserSchema

logger = logging.getLogger("picsellia")


class User(Dao):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def username(self) -> str:
        """Username of the Worker"""
        return self._username

    def __str__(self):
        return f"User '{self.username}'"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        logger.warning(
            "This method is not calling platform to refresh data. You should not call it"
        )
        return {"id": self.id, "username": self.username}

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> UserSchema:
        schema = UserSchema(**data)
        self._username = schema.username
        return schema



import logging
import os
from abc import ABC, abstractmethod
from pathlib import Path
from uuid import UUID

import requests
from beartype import beartype

from picsellia.decorators import exception_handler
from picsellia.exceptions import NoUrlAvailable
from picsellia.sdk.connexion import Connection

logger = logging.getLogger("picsellia")


class Downloadable(ABC):
    def __init__(self):
        if not hasattr(self, "_url"):
            self._url = None

    @property
    @abstractmethod
    def connexion(self) -> Connection:
        pass

    @property
    def url(self) -> str:
        """Url
        This is generated by backend and expires after 1 hour. So this property might be out of date.
        Calling `sync()` method will retrieve a new url when expired.
        """
        if not self._url:
            self.reset_url()
        return self._url

    @property
    @abstractmethod
    def id(self) -> UUID:
        """Id"""
        pass

    @property
    @abstractmethod
    def object_name(self) -> str:
        """Object name"""
        pass

    @property
    @abstractmethod
    def filename(self) -> str:
        """Filename"""
        pass

    @property
    def id_with_extension(self) -> str:
        extension = os.path.splitext(self.filename)[1]
        return f"{self.id}{extension}"

    @property
    @abstractmethod
    def large(self) -> bool:
        """If true, this (Object) has a large size"""
        pass

    @abstractmethod
    def reset_url(self) -> str:
        """Reset url of this object"""
        pass

    @exception_handler
    @beartype
    def download(
        self,
        target_path: str | Path = "./",
        force_replace: bool = False,
        use_id: bool = False,
    ) -> None:
        """Download this object into given target_path

        Examples:
            ```python
            data = clt.get_datalake().list_data(limit=1)
            data.download('./data/')
            ```

        Arguments:
            target_path (str, optional): Target path where data will be downloaded. Defaults to './'.
            force_replace (bool, optional): Replace an existing file if exists. Defaults to False.
            use_id (bool, optional): If true, will download file with id and extension as file name. Defaults to False.
        """
        if self._do_download(target_path, force_replace, use_id=use_id):
            logger.info(f"{self.filename} downloaded successfully.")
        else:
            logger.info(f"{self.filename} was already downloaded.")

    @exception_handler
    @beartype
    def _do_download(
        self,
        target_path: str | Path,
        force_replace: bool,
        use_id: bool = False,
    ) -> bool:
        if use_id:
            path = os.path.join(target_path, self.id_with_extension)
        else:
            path = os.path.join(target_path, self.filename)
        url = self.url
        if not url:
            raise NoUrlAvailable(
                "Could not retrieve an url to download this file: you might not have rights to download it."
            )

        try:
            return self.connexion.do_download_file(
                path, url, is_large=self.large, force_replace=force_replace
            )
        except requests.exceptions.HTTPError as e:
            # Handle presigned url expired
            if (
                e.response is None
                or e.response.status_code is None
                or e.response.status_code != 403
            ):
                raise

            logger.debug("url might have expired, regenerating it")
            url = self.reset_url()
            if not url:
                raise NoUrlAvailable(
                    "Could not retrieve an url to download this file: you might not have rights to download it."
                )

            return self.connexion.do_download_file(
                path, url, is_large=self.large, force_replace=force_replace
            )



import abc
import logging
from functools import partial
from typing import Any, Generic, TypeVar
from uuid import UUID

from orjson import orjson
from pydantic import BaseModel, Field

from picsellia import pxl_multithreading as mlt
from picsellia.compatibility import add_data_mandatory_query_parameters
from picsellia.pxl_multithreading import do_paginate
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.utils import flatten

logger = logging.getLogger(__name__)


class BaseFilter(BaseModel):
    limit: int | None = Field(default=None, gt=0)
    offset: int | None = Field(default=None, ge=0)
    order_by: list[str] | None = None
    query: str | None = None

    def has_list_of_items(self) -> bool:
        pass


TItem = TypeVar("TItem", bound=Dao)
TFilter = TypeVar("TFilter", bound=BaseFilter)


class AbstractItemLister(abc.ABC, Generic[TItem, TFilter]):
    def __init__(self, connexion: Connection):
        self.connexion = connexion

    def list_items(self, filters: TFilter) -> list[TItem]:
        if filters.has_list_of_items():
            return self._chunk_filter_and_fetch_items(filters)
        else:
            return self._list_items_with_pagination(filters)

    def list_ids(
        self,
        limit: int | None = None,
        offset: int | None = None,
        query: str | None = None,
    ) -> list[UUID]:
        return do_paginate(
            limit,
            offset,
            10000,
            partial(self._fetch_paginated_ids, query),
        )

    @abc.abstractmethod
    def _get_query_param_and_items_from_filters(
        self, filters: TFilter
    ) -> tuple[str, list]:
        pass

    @abc.abstractmethod
    def _get_other_params(self, filters: TFilter) -> dict[str, Any]:
        pass

    @abc.abstractmethod
    def _get_path(self) -> str:
        pass

    @abc.abstractmethod
    def _build_item(self, data: dict) -> TItem:
        pass

    def _chunk_filter_and_fetch_items(self, filters: TFilter) -> list[TItem]:
        query_param, items = self._get_query_param_and_items_from_filters(filters)
        base_params = self._get_other_params(filters)

        paginate_fn = partial(
            self._call_xget_for_one_chunk,
            filters.query,
            filters.order_by,
            base_params,
            query_param,
        )

        return flatten(
            mlt.do_chunk_called_function(
                items,
                f=paginate_fn,
                chunk_size=1000,
            )
        )

    def _call_xget_for_one_chunk(
        self,
        query: str | None,
        order_by: list[str] | None,
        base_params: dict,
        query_param: str,
        items: list,
    ) -> list[TItem]:
        params = base_params.copy()
        if order_by:
            params["order_by"] = order_by

        payload: dict[str, Any] = {query_param: items}
        if query:
            payload["q"] = query

        return self._call_xget(params, payload)[0]

    def _call_xget(self, params: dict, payload: dict) -> tuple[list[TItem], int]:
        add_data_mandatory_query_parameters(payload)
        content = self.connexion.xget(
            self._get_path(),
            params=params,
            data=orjson.dumps(payload),
        ).json()
        return self._parse_response(content)

    def _list_items_with_pagination(self, filters: TFilter) -> list[TItem]:
        base_params = self._get_other_params(filters)
        paginate_fn = partial(
            self._call_get_for_one_page, filters.query, filters.order_by, base_params
        )
        return mlt.do_paginate(filters.limit, filters.offset, None, paginate_fn)

    def _call_get_for_one_page(
        self,
        query: str | None,
        order_by: list[str] | None,
        base_params: dict[str, Any],
        limit: int,
        offset: int,
    ) -> tuple[list[TItem], int]:
        params = base_params.copy()
        params["limit"] = limit
        params["offset"] = offset
        if order_by:
            params["order_by"] = order_by
        if query:
            params["q"] = query
        return self._call_get(params)

    def _call_get(self, params: dict) -> tuple[list[TItem], int]:
        add_data_mandatory_query_parameters(params)
        content = self.connexion.get(self._get_path(), params=params).json()
        return self._parse_response(content)

    def _parse_response(self, content: dict) -> tuple[list[TItem], int]:
        return (
            [self._build_item(item) for item in content["items"]],
            content["count"],
        )

    def _fetch_paginated_ids(self, query: str | None, limit: int, offset: int):
        params: dict[str, Any] = {"limit": limit, "offset": offset}
        if query:
            params["q"] = query
        path = self._get_path() + "/ids/paginated"
        r = self.connexion.get(path, params=params).json()
        return r["items"], r["count"]



import logging
import os
from pathlib import Path
from time import sleep

from beartype import beartype
from orjson import orjson

from picsellia import exceptions, utils
from picsellia.decorators import exception_handler
from picsellia.exceptions import (
    IllegalJobTransitionError,
    NotSupportedJobVersionError,
    WaitingAttemptsTimeout,
    WrongJobVersionError,
)
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.enums import JobRunStatus, JobStatus, ObjectDataType
from picsellia.types.schemas import JobRunSchema, JobSchema

logger = logging.getLogger("picsellia")


class Job(Dao):
    def __init__(self, connexion: Connection, data: dict, version: int) -> None:
        self._version = version
        Dao.__init__(self, connexion, data)

    @property
    def status(self) -> JobStatus | JobRunStatus:
        """Status of this (Job)"""
        return self._status

    def __str__(self):
        return f"JobV{self._version} {self.id}: {self.status}"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        if self._version == 1:
            r = self.connexion.get(f"/api/job/{self.id}").json()
        elif self._version == 2:
            r = self.connexion.get(f"/api/v2/job/{self.id}").json()
        else:  # pragma: no cover
            raise NotSupportedJobVersionError(
                f"Unsupported Job version {self._version}"
            )
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> JobSchema | JobRunSchema:
        if self._version == 1:
            schema = JobSchema(**data)
        elif self._version == 2:
            schema = JobRunSchema(**data)
        else:  # pragma: no cover
            raise NotSupportedJobVersionError(
                f"Unsupported Job version {self._version}"
            )
        self._status = schema.status
        return schema

    @exception_handler
    @beartype
    def wait_for_done(self, blocking_time_increment: float = 1.0, attempts: int = 20):
        """
        Wait for the job to be done.

        Examples:
            ```python
            job.wait_for_done()
            ```

        Arguments:
            blocking_time_increment: Time between each attempts
            attempts: Number of attempts
        """
        if self._version == 1:
            statuses = [
                JobStatus.SUCCESS,
                JobStatus.FAILED,
                JobStatus.TERMINATED,
            ]
        elif self._version == 2:
            statuses = [
                JobRunStatus.SUCCEEDED,
                JobRunStatus.FAILED,
                JobRunStatus.KILLED,
            ]
        else:  # pragma: no cover
            raise NotSupportedJobVersionError(
                f"Unsupported Job version {self._version}"
            )
        return self.wait_for_status(
            statuses,
            blocking_time_increment,
            attempts,
        )

    @exception_handler
    @beartype
    def wait_for_status(
        self,
        statuses: str | JobStatus | JobRunStatus | list[str | JobStatus | JobRunStatus],
        blocking_time_increment: float = 1.0,
        attempts: int = 20,
    ) -> JobStatus | JobRunStatus:
        """
        Wait for the job to be in a specific status.

        Examples:
            ```python
            job = client.get_job_by_id("job_id")
            job.wait_for_status(JobStatus.SUCCESS)
            ```

        Arguments:
            statuses: Status to wait for (JobStatus or JobRunStatus)
            blocking_time_increment: Time between each attempts
            attempts: Number of attempts

        Returns:
            (JobStatus) or (JobRunStatus) of the job

        """
        if (
            isinstance(statuses, JobRunStatus)
            or isinstance(statuses, JobStatus)
            or isinstance(statuses, str)
        ):
            statuses = [statuses]

        if self._version == 1:
            waited_statuses = [JobStatus.validate(status) for status in statuses]
        elif self._version == 2:
            waited_statuses = [JobRunStatus.validate(status) for status in statuses]
        else:  # pragma: no cover
            raise NotSupportedJobVersionError(
                f"Unsupported Job version {self._version}"
            )

        attempt = 0
        while attempt < attempts:
            self.sync()
            if self.status in waited_statuses:
                break

            sleep(blocking_time_increment)
            attempt += 1

        if attempt >= attempts:
            raise WaitingAttemptsTimeout(
                f"Job is still not in the status you've been waiting for, after {attempt} attempts."
                "Please wait a few more moment, or check"
            )

        return self.status

    @exception_handler
    @beartype
    def update_job_run_with_status(self, status: JobRunStatus):
        """
        Update the job run with a new status.

        Examples:
            ```python
            job.update_job_run_with_status(JobRunStatus.SUCCEEDED)
            ```

        Arguments:
            status: New status of the job run
        """
        if self._version != 2:  # pragma: no cover
            raise WrongJobVersionError("You can't use this method with this job")

        if status == JobRunStatus.SUCCEEDED:
            r = self.connexion.post(f"/api/v2/job/{self.id}/succeed").json()
        elif status == JobRunStatus.FAILED:
            r = self.connexion.post(f"/api/v2/job/{self.id}/fail").json()
        elif status == JobRunStatus.RUNNING:
            r = self.connexion.post(f"/api/v2/job/{self.id}/running").json()
        else:
            raise IllegalJobTransitionError(
                f"You can change this job's status to {status}"
            )

        self.refresh(r)

    @exception_handler
    @beartype
    def send_logging(
        self,
        log: str | list,
        part: str,
        final: bool = False,
        special: str | bool | list = False,
    ) -> None:
        """Send a log entry to the job.

        Examples:
            ```python
            job.send_logging("log1", "part1")
            job.send_logging("log2", "part1")
            job.send_logging("log3", "part2")
            job.send_logging("log4", "part2", final=True)
            ```

        Arguments:
            log (str): Log content
            part (str): Logging Part
            final (bool, optional): True if Final line. Defaults to False.
            special (bool, optional): True if special log. Defaults to False.

        """
        if self._version != 2:  # pragma: no cover
            raise WrongJobVersionError("You can't use this method with this job")

        if not hasattr(self, "line_nb"):
            self.line_nb = 0

        to_send = {
            "line_nb": self.line_nb,
            "log": log,
            "final": final,
            "part": part,
            "special": special,
        }
        self.line_nb += 1
        self.connexion.post(
            f"/api/v2/job/{self.id}/logging",
            data=orjson.dumps(to_send),
        )

    @exception_handler
    @beartype
    def start_logging_chapter(self, name: str) -> None:
        """Start a logging chapter.

        Examples:
            ```python
            job.start_logging_chapter("chapter1")
            job.send_logging("log1")
            job.send_logging("log2")
            job.end_logging_chapter()
            ```

        Arguments:
            name (str): Chapter name
        """
        utils.print_start_chapter_name(name)
        utils.print_line_return()

    @exception_handler
    @beartype
    def start_logging_buffer(self, length: int = 1) -> None:
        """Start a logging buffer.

        Examples:
            ```python
            job.start_logging_buffer()
            job.send_logging("log1")
            job.send_logging("log2")
            job.end_logging_buffer()
            ```

        Arguments:
            length (int, optional): Buffer length. Defaults to 1.
        """
        utils.print_logging_buffer(length)
        self.buffer_length = length

    @exception_handler
    @beartype
    def end_logging_buffer(self) -> None:
        """End the logging buffer."""
        utils.print_logging_buffer(self.buffer_length)

    @exception_handler
    @beartype
    def store_logging_file(self, path: str | Path) -> str:
        """Store a logging file in the platform.

        Examples:
            ```python
            job.store_logging_file("path/to/file")
            ```
        Arguments:
            path (str or Path): path to the file or folder.

        Raises:
            FileNotFoundException: No file found at the given path

        Returns:
            Object name of logging file as string
        """
        if self._version != 2:  # pragma: no cover
            raise WrongJobVersionError("You can't use this method with this job")

        if not os.path.exists(path):
            raise exceptions.FileNotFoundException(f"{path} not found")

        filename = os.path.basename(path)
        object_name = self.connexion.generate_job_object_name(
            filename, ObjectDataType.LOGGING, self.id
        )
        _, is_large, _ = self.connexion.upload_file(object_name, path)

        payload = {"logging_object_name": object_name}

        r = self.connexion.post(
            f"/api/v2/job/{self.id}/logging/save",
            data=orjson.dumps(payload),
        ).json()
        return r["logging_object_name"]



import logging
from uuid import UUID

import orjson
from beartype import beartype

import picsellia.exceptions as exceptions
from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.sdk.data import Data
from picsellia.sdk.downloadable import Downloadable
from picsellia.sdk.multi_object import MultiObject
from picsellia.types.enums import DataType
from picsellia.types.schemas import PredictedAssetSchema

logger = logging.getLogger("picsellia")


class PredictedAsset(Dao, Downloadable):
    def __init__(self, connexion: Connection, deployment_id: UUID, data: dict):
        Dao.__init__(self, connexion, data)
        Downloadable.__init__(self)

        self._deployment_id = deployment_id

    @property
    def deployment_id(self) -> UUID:
        """UUID of (Deployment) where this (PredictedAsset) is"""
        return self._deployment_id

    @property
    def data_id(self) -> UUID:
        """UUID of (Data) of this (PredictedAsset)"""
        return self._data_id

    @property
    def oracle_prediction_id(self) -> UUID:
        """Oracle prediction if of this predicted asset"""
        return self._oracle_prediction_id

    @property
    def object_name(self) -> str:
        """Object name of this (PredictedAsset)"""
        return self._object_name

    @property
    def filename(self) -> str:
        """Filename of this (PredictedAsset)"""
        return self._filename

    @property
    def large(self) -> bool:
        """If true, this (Asset) file is considered large"""
        return True

    @property
    def type(self) -> DataType:
        """Type of this (PredictedAsset)"""
        return self._type

    @property
    def width(self) -> int:
        """Width of this (PredictedAsset)."""
        return self._width

    @property
    def height(self) -> int:
        """Height of this (PredictedAsset)."""
        return self._height

    @property
    def metadata(self) -> dict | None:
        """Metadata of this Data. Can be None"""
        return self._metadata

    def __str__(self):
        return f"{Colors.BLUE}PredictedAsset '{self.filename}' ({self.type}) {Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def refresh(self, data: dict) -> PredictedAssetSchema:
        schema = PredictedAssetSchema(**data)
        self._data_id = schema.data.id
        self._oracle_prediction_id = schema.oracle_prediction_id

        # Downloadable properties
        self._object_name = schema.data.object_name
        self._filename = schema.data.filename
        self._url = schema.data.url
        self._metadata = schema.data.metadata

        self._type = schema.data.type
        self._height = schema.data.meta.height
        self._width = schema.data.meta.width

        self._champion_prediction = None
        self._shadow_prediction = None
        self._review = None
        return schema

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/predicted/asset/{self.id}").json()
        self.refresh(r)
        self._load_predictions(r["predictions"], r["reviews"])
        return r

    @exception_handler
    @beartype
    def reset_url(self) -> str:
        """Reset url property of this Asset by calling platform.

        Returns:
            A url as str of this Asset.
        """
        r = self.connexion.get(f"/api/data/{self.data_id}/presigned-url")
        self._url = r.json()["presigned_url"]
        return self._url

    @exception_handler
    @beartype
    def get_data(self):
        """Retrieve data of this asset

            ```python
            data = asset.get_data()
            assert data.id == asset.data_id
            assert data.filename == asset.filename
            ```

        Returns:
            A (Data) object
        """
        r = self.connexion.get(f"/api/data/{self.data_id}").json()
        return Data(self.connexion, UUID(r["datalake_id"]), r)

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this predicted asset from its deployment

        :warning: **DANGER ZONE**: Be very careful here!

        Remove this predicted asset

        Examples:
            ```python
            one_asset.delete()
            ```
        """
        self.connexion.delete(f"/api/predicted/asset/{self.id}")
        logger.info(f"{self} removed from deployment")

    @exception_handler
    @beartype
    def add_review(
        self,
        rectangles: list[tuple[int, int, int, int, str]] | None = None,
        polygons: list[tuple[list[list[int]], str]] | None = None,
        classifications: list[str] | None = None,
        lines: list[tuple[list[list[int]], str]] | None = None,
        points: list[tuple[list[int], str]] | None = None,
    ):
        """Add a review to your predicted-asset.
           It will then be used along the prediction for the monitoring.

        :warning: **DANGER ZONE**: You will not be able to change it afterward and
            the monitoring metrics will not be able to be computed again with another review.


        Examples:
            ```python
            one_annotation.overwrite(rectangles=[(10, 20, 30, 40, label_cat), (50, 60, 20, 30, label_dog)])
            ```

        Arguments:
            rectangles (list[tuple[int, int, int, int, str]], optional): List of rectangles of this review. Defaults to None.
            polygons (list[tuple[list[list[int]], str]], optional): List of polygons of this review. Defaults to None.
            classifications (list[str], optional): List of classifications of this review. Defaults to None.
            lines (list[tuple[list[list[int]], str]], optional): List of lines of this review. Defaults to None.
            points (list[tuple[list[int], str]], optional): List of points of this review. Defaults to None.
        """
        payload = {}
        if rectangles:
            payload["rectangles"] = [
                {
                    "x": rectangle[0],
                    "y": rectangle[1],
                    "w": rectangle[2],
                    "h": rectangle[3],
                    "label": rectangle[4],
                }
                for rectangle in rectangles
            ]

        if polygons:
            payload["polygons"] = [
                {"polygon": polygon[0], "label": polygon[1]} for polygon in polygons
            ]
        if classifications:
            payload["classifications"] = [
                {"label": classification} for classification in classifications
            ]
        if lines:
            payload["lines"] = [{"line": line[0], "label": line[1]} for line in lines]
        if points:
            payload["points"] = [
                {"point": point[0], "label": point[1]} for point in points
            ]
        r = self.connexion.post(
            f"/api/predicted/asset/{self.id}/reviews", data=orjson.dumps(payload)
        ).json()
        logger.info(f"{self} has been reviewed.")
        return r

    @exception_handler
    @beartype
    def get_champion_prediction(self) -> dict:
        """This will return a dict with data representing champion prediction of this asset"""
        if self._champion_prediction:
            return self._champion_prediction

        data = self.sync()
        self._load_predictions(data["predictions"], data["reviews"])
        return self._champion_prediction

    @exception_handler
    @beartype
    def get_shadow_prediction(self) -> dict | None:
        """This will return a dict with data representing shadow prediction of this asset"""
        if self._shadow_prediction:
            return self._shadow_prediction

        data = self.sync()
        self._load_predictions(data["predictions"], data["reviews"])
        return self._shadow_prediction

    @exception_handler
    @beartype
    def get_review(self) -> dict | None:
        """This will return a dict with data representing the last review created on this asset"""
        if self._review:
            return self._review

        data = self.sync()
        self._load_predictions(data["predictions"], data["reviews"])
        return self._review

    def _load_predictions(
        self, predictions: list[dict] | None, reviews: list[dict] | None
    ) -> None:
        if predictions:
            for prediction in predictions:
                if prediction["is_shadow"]:
                    self._shadow_prediction = prediction
                else:
                    self._champion_prediction = prediction

        if reviews:
            self._review = reviews[-1]


class MultiPredictedAsset(MultiObject[PredictedAsset]):
    @beartype
    def __init__(
        self,
        connexion: Connection,
        deployment_id: UUID,
        items: list[PredictedAsset],
    ):
        MultiObject.__init__(self, connexion, items)
        self.deployment_id = deployment_id

    def __str__(self) -> str:
        return f"{Colors.GREEN}MultiPredictedAsset for deployment {self.deployment_id} {Colors.ENDC}, size: {len(self)}"

    def __getitem__(self, key) -> "PredictedAsset | MultiPredictedAsset":
        if isinstance(key, slice):
            indices = range(*key.indices(len(self.items)))
            assets = [self.items[i] for i in indices]
            return MultiPredictedAsset(self.connexion, self.deployment_id, assets)
        return self.items[key]

    @beartype
    def __add__(self, other) -> "MultiPredictedAsset":
        self.assert_same_connexion(other)
        items = self.items.copy()
        if isinstance(other, MultiPredictedAsset):
            items.extend(other.items.copy())
        elif isinstance(other, PredictedAsset):
            items.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

        return MultiPredictedAsset(self.connexion, self.deployment_id, items)

    @beartype
    def __iadd__(self, other) -> "MultiPredictedAsset":
        self.assert_same_connexion(other)

        if isinstance(other, MultiPredictedAsset):
            self.extend(other.items.copy())
        elif isinstance(other, PredictedAsset):
            self.append(other)
        else:
            raise exceptions.BadRequestError("You can't add these two objects")

        return self



from uuid import UUID

from picsellia.exceptions import NoDataError
from picsellia.sdk.artifact import Artifact
from picsellia.sdk.connexion import Connection
from picsellia.sdk.evaluation import MultiEvaluation
from picsellia.sdk.log import Log
from picsellia.services.lister.evaluation_lister import (
    EvaluationFilter,
    EvaluationLister,
)


class TrainingService:
    """This service is used by Experiment and FastTraining dao to fetch evaluations, artifacts and logs."""

    def __init__(self, connexion: Connection, experiment_id: UUID):
        self.connexion = connexion
        self.experiment_id = experiment_id

    def list_evaluations(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
        q: str | None = None,
    ) -> MultiEvaluation:
        filters = EvaluationFilter.model_validate(
            {"limit": limit, "offset": offset, "order_by": order_by, "query": q}
        )
        evaluations = EvaluationLister(self.connexion, self.experiment_id).list_items(
            filters
        )

        if len(evaluations) == 0:
            raise NoDataError(
                "No evaluation done by this training found with this query"
            )

        return MultiEvaluation(self.connexion, self.experiment_id, evaluations)

    def list_logs(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Log]:
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        r = self.connexion.get(
            f"/api/experiment/{self.experiment_id}/logs/extended", params=params
        ).json()
        return [Log(self.connexion, item) for item in r["items"]]

    def get_log(self, name: str) -> Log:
        params = {"name": name}
        r = self.connexion.get(
            f"/api/experiment/{self.experiment_id}/logs/find", params=params
        ).json()
        return Log(self.connexion, r)

    def list_artifacts(
        self,
        limit: int | None = None,
        offset: int | None = None,
        order_by: list[str] | None = None,
    ) -> list[Artifact]:
        params = {"limit": limit, "offset": offset, "order_by": order_by}
        r = self.connexion.get(
            f"/api/experiment/{self.experiment_id}/artifacts", params=params
        ).json()
        return [Artifact(self.connexion, item) for item in r["items"]]

    def get_parameters(self) -> dict:
        return self.get_log("parameters").data



import logging
from uuid import UUID

from picsellia.sdk.user import User
from picsellia.sdk.worker import Worker
from picsellia.types.enums import WorkerType

logger = logging.getLogger("picsellia")


def parse_user_weights_to_assignees(
    assignees: list[tuple[Worker | User | UUID, float | int]],
    expected_worker_type: WorkerType | None = None,
):
    users = []
    for user, weight in assignees:
        if isinstance(user, Worker):
            logger.warning(
                "You should not use Worker anymore, instead use User or UUID for an user id"
            )
            if expected_worker_type and user.type != expected_worker_type:
                raise ValueError(f"Assignee {user} cannot be used for this campaign.")
            user_id = user.user_id
        elif isinstance(user, User):
            user_id = user.id
        else:
            # Assume it's a user id and not a worker_id anymore
            user_id = user
        users.append({"user_id": user_id, "weight": weight})

    return users



from abc import ABC, abstractmethod
from uuid import UUID

from picsellia.exceptions import UndefinedObjectError
from picsellia.sdk.connexion import Connection
from picsellia.types.schemas import DaoSchema


class Dao(ABC):
    def __init__(self, connexion: Connection, data: dict) -> None:
        self._connexion = connexion
        schema: DaoSchema = self.refresh(data)
        if schema.id is None:  # pragma: no cover
            raise UndefinedObjectError(
                "This object has no id. Something went wrong while retrieving data from platform."
            )
        self._id = schema.id

    @property
    def connexion(self) -> Connection:
        """Connection object used to connect to Picsellia platform"""
        return self._connexion

    @property
    def id(self) -> UUID:
        """UUID in Picsellia platform"""
        return self._id

    @property
    def ids(self) -> list[UUID]:
        return [self._id]

    def __str__(self) -> str:
        return f"Platform object (id: {self.id})"

    def __repr__(self) -> str:
        return self.__str__()

    def __eq__(self, __o: object) -> bool:
        if isinstance(__o, type(self)):
            __o: Dao
            return self._id == __o.id and self._connexion == __o.connexion

        return False

    @abstractmethod
    def refresh(self, data: dict) -> DaoSchema:
        """
        Refresh the properties of this object with given data.
        This will not update the object in Picsellia database.
        It is only useful in some case, if you want to manually override attributes of this object.
        """
        pass

    @abstractmethod
    def sync(self) -> dict:
        """
        Retrieve the data of this object and reset properties.
        Useful when objects were retrieved a long time ago and to ensures synchronization with Picsellia database.
        This will also return some data stored in database, not used in SDK, as a dict.
        """
        pass



import logging

import orjson
from beartype import beartype

from picsellia.colors import Colors
from picsellia.decorators import exception_handler
from picsellia.exceptions import BadRequestError
from picsellia.sdk.connexion import Connection
from picsellia.sdk.dao import Dao
from picsellia.types.enums import LogType
from picsellia.types.schemas import LogDataType, LogSchema
from picsellia.utils import filter_payload

logger = logging.getLogger("picsellia")


class Log(Dao):
    def __init__(self, connexion: Connection, data: dict):
        Dao.__init__(self, connexion, data)

    @property
    def name(self) -> str:
        """Name of this (Log)"""
        return self._name

    @property
    def data(self) -> LogDataType:
        """Data of this (Log)"""
        return self._data

    @property
    def type(self) -> LogType:
        """Type of this (Log)"""
        return self._type

    def __str__(self):
        return f"{Colors.GREEN}Log {self.name}{Colors.ENDC} (id: {self.id})"

    @exception_handler
    @beartype
    def sync(self) -> dict:
        r = self.connexion.get(f"/api/log/{self.id}").json()
        self.refresh(r)
        return r

    @exception_handler
    @beartype
    def refresh(self, data: dict):
        schema = LogSchema(**data)
        self._name = schema.name
        self._data = schema.data
        self._type = schema.type
        return schema

    @exception_handler
    @beartype
    def update(self, name: str | None = None, data: LogDataType | None = None) -> None:
        """Update this log with a new name or new data

        You cannot change the type of this Log.

        Examples:
            ```python
            my_log.update(name="new_name", data={"key": "value"})
            ```
        Arguments:
            name (str, optional): New name of the log. Defaults to None.
            data (LogDataType, optional): New data of the log. Defaults to None.
        """
        if self.type == LogType.IMAGE and data:
            raise BadRequestError(
                "You cannot update data of a log image this way, use experiment.log() method instead."
            )

        payload = {"name": name, "data": data}
        filtered_payload = filter_payload(payload)
        r = self.connexion.patch(
            f"/api/log/{self.id}", data=orjson.dumps(filtered_payload)
        ).json()
        self.refresh(r)
        logger.info(f"{self} updated.")

    @exception_handler
    @beartype
    def delete(self) -> None:
        """Delete this log

        Examples:
            ```python
            my_log.delete()
            ```
        """
        self.connexion.delete(f"/api/log/{self.id}")

    @exception_handler
    @beartype
    def append(self, data: LogDataType) -> None:
        """Appends value to log with given name.

        You can only append log on Line logs.

        Examples:
            ```python
            my_log.append(data={"key": "value"})
            ```

        Arguments:
            data (LogDataType): Data to append to this log
        """
        assert self.type == LogType.LINE, "You can only append log on Line logs"
        payload = {"data": data}
        r = self.connexion.post(
            f"/api/log/{self.id}/append", data=orjson.dumps(payload)
        ).json()
        self.refresh(r)
        logger.info(f"Append {data} on {self}.")



from picsellia.types.enums import DataType, DataUploadStatus


def add_data_mandatory_query_parameters(params: dict):
    # As we can now handle processing data and video, we can fetch all data
    params["upload_statuses"] = [
        DataUploadStatus.PENDING.value,
        DataUploadStatus.COMPUTING.value,
        DataUploadStatus.DONE.value,
        DataUploadStatus.ERROR.value,
    ]
    params["types"] = [DataType.IMAGE.value, DataType.VIDEO.value]
    return params
